"""Polarizability model based on atomic Raman tensors (ARTs)."""

from __future__ import annotations
from typing import cast
from pathlib import Path

import numpy as np
from numpy.typing import NDArray
from tabulate import tabulate

from ramannoodle.constants import ANSICOLORS
from ramannoodle.pmodel._interpolation import InterpolationModel
from ramannoodle.exceptions import (
    get_type_error,
    get_shape_error,
    verify_ndarray_shape,
    InvalidDOFException,
    UserError,
)

PYMATGEN_PRESENT = True
try:
    import pymatgen.core
except (UserError, ImportError):
    PYMATGEN_PRESENT = False


def _get_directions_str(directions: list[NDArray[np.float64]]) -> str:
    """Return easy-to-read string for directions."""
    direction_strs = []
    for direction in directions:
        direction_strs.append(
            np.array2string(direction, precision=5, suppress_small=True, sign="+")
        )
    return ", ".join(direction_strs)


def _get_specified_str(num_specified: int) -> str:
    """Return ansi-colored string for number of specified models."""
    core = f"{num_specified}/3"
    if num_specified == 3:
        return ANSICOLORS.OK_GREEN + core + ANSICOLORS.END
    if 1 <= num_specified <= 2:
        return ANSICOLORS.WARNING_YELLOW + core + ANSICOLORS.END
    return ANSICOLORS.ERROR_RED + core + ANSICOLORS.END


class ARTModel(InterpolationModel):
    """Polarizability model based on atomic Raman tensors (ARTs).

    This model is a flavor of :class:`.InterpolationModel` with added restrictions. The
    degrees of freedom (DOF) in ARTModel are displacements of individual atoms. 1D
    interpolations are used, with only two amplitudes allowed per DOF. Atomic
    displacements can be in any direction, so long as all three directions for a given
    atom are orthogonal.

    All assumptions from :class:`.InterpolationModel` apply.

    .. note::
        Degrees of freedom cannot (and should not) be added using the :meth:`.add_dof`
        and :meth:`.add_dof_from_files` methods inherited from
        :class:`.InterpolationModel`. Usage of these methods will raise a
        :class:`.UserError`. Instead, use :meth:`add_art` or
        :meth:`add_art_from_files`.

    .. warning::
        Due to the way ARTModel is implemented, the predicted tensor of the reference
        structure may be offset slightly. This is a fixed offset and therefore has no
        influence on Raman spectra calculations.

    Parameters
    ----------
    ref_structure
        Reference structure on which to base the model.
    ref_polarizability
        Array with shape (3,3) giving polarizability of the reference structure.
    is_dummy_model

    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def add_dof(  # pylint: disable=too-many-arguments
        self,
        cart_displacement: NDArray[np.float64],
        amplitudes: NDArray[np.float64],
        polarizabilities: NDArray[np.float64],
        interpolation_order: int,
        include_ref_polarizability: bool = True,
    ) -> None:
        """Disable add_dof.

        Raises
        ------
        UserError

        :meta private:
        """
        raise UserError("add_dof should not be used; use add_art instead")

    def add_dof_from_files(
        self,
        filepaths: str | Path | list[str] | list[Path],
        file_format: str,
        interpolation_order: int,
    ) -> None:
        """Disable add_dof_from_files.

        Raises
        ------
        UserError

        :meta private:
        """
        raise UserError(
            "add_dof_from_files should not be used; use add_art_from_files instead"
        )

    def add_dof_from_pymatgen(
        self,
        pymatgen_structures: "list[pymatgen.core.Structure]",
        polarizabilities: NDArray[np.float64],
        interpolation_order: int,
    ) -> None:
        """Disable add_dof_from_pymatgen.

        Raises
        ------
        UserError

        :meta private:
        """
        raise UserError(
            "add_dof_from_pymatgen should not be used; use add_art_from_pymatgen "
            "instead"
        )

    def add_art(
        self,
        atom_index: int,
        cart_direction: NDArray[np.float64],
        amplitudes: NDArray[np.float64],
        polarizabilities: NDArray[np.float64],
    ) -> None:
        """Add an atomic Raman tensor (ART).

        Specification of an ART requires an atom index, displacement direction,
        displacement amplitudes, and corresponding known polarizabilities for each
        amplitude. ARTs related to the specified ART by symmetry are added
        automatically.

        Parameters
        ----------
        atom_index
            Atom index in the reference structure.
        cart_direction
            (Å) Array with shape (3,). Must be orthogonal to all previously added
            ARTs belonging to the atom at ``atom_index``.
        amplitudes
            (Å) Array with shape (1,) or (2,). Duplicate amplitudes, either those
            explicitly provided or those generated by structural symmetries, will
            raise :class:`.InvalidDOFException`.
        polarizabilities
            Array with shape (1,3,3) or (2,3,3) containing known
            polarizabilities for each amplitude. If dummy model, this parameter is
            ignored.

        Raises
        ------
        InvalidDOFException
            Provided ART was invalid.

        """
        direction = self.ref_structure.get_frac_direction(cart_direction)
        if not isinstance(atom_index, int):
            raise get_type_error("atom_index", atom_index, "int")
        try:
            if amplitudes.shape not in [(1,), (2,)]:
                raise get_shape_error("amplitudes", amplitudes, "(1,) or (2,)")
        except AttributeError as exc:
            raise get_type_error("amplitudes", amplitudes, "ndarray") from exc
        if self._is_dummy_model:
            polarizabilities = np.zeros((amplitudes.size, 3, 3))
        verify_ndarray_shape(
            "polarizabilities", polarizabilities, (amplitudes.size, 3, 3)
        )

        displacement = self._ref_structure.positions * 0
        displacement[atom_index] = direction / np.linalg.norm(direction * 10.0)

        # Checks amplitudes and displacements
        super().add_dof(
            self.ref_structure.get_cart_displacement(displacement),
            amplitudes,
            polarizabilities,
            1,
            include_ref_polarizability=False,
        )

    def add_art_from_files(
        self,
        filepaths: str | Path | list[str] | list[Path],
        file_format: str,
    ) -> None:
        """Add an atomic Raman tensor (ART) from file(s).

        Required directions, amplitudes, and polarizabilities are automatically
        determined from provided files. Files should be chosen such that the
        resulting ARTs are valid under the same restrictions of :meth:`add_art`.

        Parameters
        ----------
        filepaths
        file_format
            Supports ``"outcar"`` and ``"vasprun.xml"``. If dummy model, supports
            ``"poscar"`` and ``"xdatcar"`` as well (see :ref:`Supported formats`).

        Raises
        ------
        FileNotFoundError
            File not found.
        InvalidFileException
            Invalid file.
        InvalidDOFException
            ART assembled from supplied files was invalid. See :meth:`add_art` for
            restrictions.

        """
        displacements, amplitudes, polarizabilities = super()._read_dof(
            filepaths, file_format
        )
        # Check whether only one atom is displaced.
        _displacement = displacements[0].copy()
        atom_index = int(np.argmax(np.sum(_displacement**2, axis=1)))
        _displacement[atom_index] = np.zeros(3)
        if not np.allclose(_displacement, 0.0, atol=1e-6):
            raise InvalidDOFException("multiple atoms displaced simultaneously")

        # Checks displacement
        basis_vectors_to_add, interpolation_xs, interpolation_ys = super()._get_dof(
            displacements[0], amplitudes, polarizabilities, False
        )

        num_amplitudes = len(interpolation_xs[0])
        if num_amplitudes != 2:
            raise InvalidDOFException(
                f"wrong number of amplitudes: {num_amplitudes} != 2"
            )

        # Checks amplitudes
        super()._construct_and_add_interpolations(
            basis_vectors_to_add, interpolation_xs, interpolation_ys, 1
        )

    def add_art_from_pymatgen(
        self,
        pymatgen_structures: list[pymatgen.core.Structure],
        polarizabilities: NDArray[np.float64],
    ) -> None:
        """
        Add an atomic Raman tensor (ART) from pymatgen Structures and polarizabilities.

        Required directions, amplitudes, and polarizabilities are automatically
        determined from provided structures. Structures should be chosen such that the
        resulting ARTs are valid under the same restrictions of :meth:`add_art`.

        Parameters
        ----------
        pymatgen_structures
            List of length M.
        polarizabilities
            Array with shape (M,3,3).

        Raises
        ------
        InvalidDOFException
            ART assembled from supplied Structures was invalid. See :meth:`add_art` for
            restrictions.
        """
        displacements, amplitudes = super()._get_displacement_amplitudes_pymatgen(
            pymatgen_structures
        )

        # Check whether only one atom is displaced.
        _displacement = displacements[0].copy()
        atom_index = int(np.argmax(np.sum(_displacement**2, axis=1)))
        _displacement[atom_index] = np.zeros(3)
        if not np.allclose(_displacement, 0.0, atol=1e-6):
            raise InvalidDOFException("multiple atoms displaced simultaneously")

        # Checks displacement
        basis_vectors_to_add, interpolation_xs, interpolation_ys = super()._get_dof(
            displacements[0], amplitudes, polarizabilities, False
        )

        num_amplitudes = len(interpolation_xs[0])
        if num_amplitudes != 2:
            raise InvalidDOFException(
                f"wrong number of amplitudes: {num_amplitudes} != 2"
            )

        # Checks amplitudes
        super()._construct_and_add_interpolations(
            basis_vectors_to_add, interpolation_xs, interpolation_ys, 1
        )

    def get_specification_tuples(
        self,
    ) -> list[tuple[int, list[int], list[NDArray[np.float64]]]]:
        """Return tuples with information on model.

        Returns
        -------
        :

            0.  parent atom index
            #.  List of atom indexes that are symmetrically equivalent to
                parent atom index
            #.  Currently specified ART directions for parent atom indexes -- (Å) List
                of arrays with shape (3,).

        """
        equivalent_atom_dict = self._ref_structure.get_equivalent_atom_dict()

        specification_tuples = []
        for atom_index in equivalent_atom_dict:
            specification_tuples.append(
                (
                    atom_index,
                    equivalent_atom_dict[atom_index],
                    self._get_art_cart_directions(atom_index),
                )
            )
        return specification_tuples

    def get_dof_indexes(
        self, atom_indexes_or_symbols: int | str | list[int | str]
    ) -> list[int]:
        """Return the DOF indexes of certain atoms.

        Parameters
        ----------
        atom_indexes_or_symbols
            If integer or list of integers, specifies atom indexes. If string or list
            of strings, specifies atom symbols. Mixtures of indexes and symbols are
            allowed.

        """
        if not isinstance(atom_indexes_or_symbols, list):
            atom_indexes_or_symbols = list([atom_indexes_or_symbols])

        atom_indexes = []
        for item in atom_indexes_or_symbols:
            if isinstance(item, str):
                atom_indexes += self._ref_structure.get_atom_indexes(item)
            else:
                atom_indexes += [item]
        atom_indexes = list(set(atom_indexes))

        dof_indexes = []
        for atom_index in atom_indexes:
            for index, basis_vector in enumerate(self._cart_basis_vectors):
                direction = basis_vector[atom_index]
                if not np.allclose(direction, 0, atol=1e-5):
                    dof_indexes.append(index)
        return dof_indexes

    def _get_art_cart_directions(self, atom_index: int) -> list[NDArray[np.float64]]:
        """Return specified art cartesian direction vectors for an atom."""
        indexes = self.get_dof_indexes(atom_index)
        cart_directions = [
            self._cart_basis_vectors[index][atom_index] for index in indexes
        ]
        return cart_directions

    def __repr__(self) -> str:
        """Get string representation."""
        specification_tuples = self.get_specification_tuples()

        table = [["Atom index", "Directions", "Specified", "Equivalent atoms"]]
        for (
            atom_index,
            equivalent_atom_indexes,
            cart_directions,
        ) in specification_tuples:
            row = [str(atom_index)]
            row.append(_get_directions_str(cart_directions))
            row.append(_get_specified_str(len(cart_directions)))
            row.append(str(len(equivalent_atom_indexes)))
            table.append(row)
        result = tabulate(table, headers="firstrow", tablefmt="rounded_outline")

        num_masked = np.sum(self._mask)
        if num_masked > 0:
            num_arts = len(self._cart_basis_vectors)
            msg = f"ATTENTION: {num_masked}/{num_arts} atomic Raman tensors are masked."
            result += f"\n {ANSICOLORS.WARNING_YELLOW} {msg} {ANSICOLORS.END}"
        if self._is_dummy_model > 0:
            msg = "ATTENTION: this is a dummy model."
            result += f"\n {ANSICOLORS.WARNING_YELLOW} {msg} {ANSICOLORS.END}"

        return result

    def get_masked_model(self, dof_indexes_to_mask: list[int]) -> ARTModel:
        """Return new model with certain degrees of freedom masked (deactivated).

        Model masking allows for the calculation of partial Raman spectra in which only
        certain degrees of freedom are considered.

        Parameters
        ----------
        dof_indexes_to_mask
            DOF indexes associated with specific atoms can be retrieved using
            :meth:`get_dof_indexes`.
        """
        # We "cast" here due to how typing is done to support Python 3.10.
        return cast(ARTModel, super().get_masked_model(dof_indexes_to_mask))



"""Utility functions relevant to symmetry."""

from typing import Iterable

import numpy as np
from numpy.typing import NDArray

from ramannoodle.exceptions import (
    get_type_error,
)


def are_collinear(vector_1: NDArray[np.float64], vector_2: NDArray[np.float64]) -> bool:
    """Check whether two vectors are collinear.

    Parameters
    ----------
    vector_1
        Array with shape (M,).
    vector_2
        Array with shape (M,).

    """
    try:
        vector_1 = vector_1 / float(np.linalg.norm(vector_1))
    except TypeError as exc:
        raise get_type_error("vector_1", vector_1, "ndarray") from exc
    try:
        vector_2 = vector_2 / float(np.linalg.norm(vector_2))
    except TypeError as exc:
        raise get_type_error("vector_2", vector_2, "ndarray") from exc
    try:
        dot_product = vector_1.dot(vector_2)
    except ValueError as exc:
        length_expr = f"{len(vector_1)} != {len(vector_2)}"
        raise ValueError(
            f"vector_1 and vector_2 have different lengths: {length_expr}"
        ) from exc
    return bool(np.allclose(dot_product, 1) or np.isclose(dot_product, -1))


def is_orthogonal_to_all(
    vector_1: NDArray[np.float64], vectors: Iterable[NDArray[np.float64]]
) -> int:
    """Check whether a vector is orthogonal to a list of other vectors.

    Parameters
    ----------
    vector_1
        Array with shape (M,).
    vectors
        Iterable containing arrays with shape (M,).

    Returns
    -------
    :
        First index of non-orthogonal vector, otherwise -1.

    """
    # This implementation could be made more efficient.
    try:
        vector_1 = vector_1 / float(np.linalg.norm(vector_1))
    except TypeError as exc:
        raise get_type_error("vector_1", vector_1, "ndarray") from exc

    for index, vector_2 in enumerate(vectors):
        try:
            vector_2 = vector_2 / np.linalg.norm(vector_2)
        except TypeError as exc:
            raise get_type_error(f"vectors[{index}]", vector_2, "ndarray") from exc

        if not np.allclose(np.dot(vector_1.flatten(), vector_2.flatten()) + 1, 1):
            return index

    return -1


def is_collinear_with_all(
    vector_1: NDArray[np.float64], vectors: Iterable[NDArray[np.float64]]
) -> int:
    """Check if a vector is collinear to a list of other vectors.

    Parameters
    ----------
    vector_1
        Array with shape (M,).
    vectors
        Iterable containing arrays with shape (M,).

    Returns
    -------
    :
        First index of non-collinear vector, otherwise -1.

    """
    # This implementation could be made more efficient.
    for index, vector_2 in enumerate(vectors):
        if not are_collinear(vector_1.flatten(), vector_2.flatten()):
            return index

    return -1


def is_non_collinear_with_all(
    vector_1: NDArray[np.float64], vectors: Iterable[NDArray[np.float64]]
) -> int:
    """Check if a vector is non-collinear to a list of other vectors.

    Parameters
    ----------
    vector_1
        Array with shape (M,).
    vectors
        Iterable containing arrays with shape (M,).

    Returns
    -------
    :
        First index of collinear vector, otherwise -1.

    """
    # This implementation could be made more efficient.
    for index, vector_2 in enumerate(vectors):
        if are_collinear(vector_1.flatten(), vector_2.flatten()):
            return index

    return -1



"""Modules for polarizability models implemented with PyTorch."""

# flake8: noqa: F401
from ramannoodle.pmodel.torch._gnn import PotGNN
from ramannoodle.pmodel.torch._train import train_single_epoch
from ramannoodle.pmodel.torch import _utils

__all__ = ["PotGNN", "train_single_epoch", "_utils"]



"""Functions for generating and writing displaced structures.

These functions are useful for preparing polarizability calculations needed for
:class:`~.InterpolationModel` and :class:`~.ARTModel`.

"""

# Design note:
# These functions are not implemented in ReferenceStructure to give
# greater modularity. For example, when different displacement methods are added (such
# as Monte Carlo rattling or random displacements), we'd rather not add code to
# Reference Structure. These functions stand alone, just like the IO functions.

from pathlib import Path

import numpy as np
from numpy.typing import NDArray

from ramannoodle.structure._reference import ReferenceStructure
from ramannoodle.structure.utils import displace_positions
from ramannoodle.exceptions import (
    get_type_error,
    get_shape_error,
    verify_ndarray_shape,
    verify_list_len,
)
from ramannoodle.io._utils import pathify_as_list
import ramannoodle.io.generic as generic_io


def get_displaced_positions(
    ref_structure: ReferenceStructure,
    cart_displacement: NDArray[np.float64],
    amplitudes: NDArray[np.float64],
) -> list[NDArray[np.float64]]:
    """Return positions displaced along a certain displacement.

    Parameters
    ----------
    ref_structure
        Reference structure containing N atoms.
    cart_displacement
        (Å) Array with shape (N,3). The magnitude of the displacement is ignored,
        only the direction is used.
    amplitudes
        (Å) Array with shape (M,).

    Returns
    -------
    :
        (fractional) List of length M containing arrays with shape (N,3).

    """
    try:
        cart_displacement = cart_displacement / float(np.linalg.norm(cart_displacement))
    except TypeError as err:
        raise get_type_error("cart_displacement", cart_displacement, "ndarray") from err
    verify_ndarray_shape("amplitudes", amplitudes, (None,))

    positions = []
    for amplitude in amplitudes:
        try:
            displacement = ref_structure.get_frac_displacement(
                cart_displacement * amplitude
            )
        except ValueError as err:
            raise get_shape_error(
                "cart_displacement",
                cart_displacement,
                f"({len(ref_structure.positions)}, 3)",
            ) from err
        positions.append(displace_positions(ref_structure.positions, displacement))

    return positions


# pylint: disable=too-many-arguments,too-many-positional-arguments
def write_displaced_structures(
    ref_structure: ReferenceStructure,
    cart_displacement: NDArray[np.float64],
    amplitudes: NDArray[np.float64],
    filepaths: str | Path | list[str] | list[Path],
    file_format: str,
    overwrite: bool = False,
) -> None:
    """Write displaced structures to files.

    Parameters
    ----------
    ref_structure
        Reference structure containing N atoms
    cart_displacement
        (Å) Array with shape (N,3). The magnitude of the displacement is ignored,
        only the direction is used.
    amplitudes
        (Å) Array with shape (M,).
    filepaths
    file_format
        Supports ``"poscar"`` (see :ref:`Supported formats`).
    overwrite
        If ``True``, overwrite the file if it exists.
    """
    filepaths = pathify_as_list(filepaths)
    position_list = get_displaced_positions(
        ref_structure, cart_displacement, amplitudes
    )
    verify_list_len("filepaths", filepaths, len(position_list))

    for position, filepath in zip(position_list, filepaths):
        generic_io.write_structure(
            ref_structure.lattice,
            ref_structure.atomic_numbers,
            position,
            filepath,
            file_format,
            overwrite,
        )


def get_ast_displaced_positions(
    ref_structure: ReferenceStructure,
    atom_index: int,
    cart_direction: NDArray[np.float64],
    amplitudes: NDArray[np.float64],
) -> list[NDArray[np.float64]]:
    """Return displaced positions with a single atom displaced along a direction.

    Parameters
    ----------
    ref_structure
        Reference structure containing N atoms.
    atom_index
    cart_direction
        (Å) Array with shape (3,).  The magnitude of the direction vector is ignored,
        only the direction is used.
    amplitudes
        (Å) Array with shape (M,).

    Returns
    -------
    :
        (fractional) List of length M containing arrays with shape (N,3).
    """
    try:
        cart_direction = cart_direction / float(np.linalg.norm(cart_direction))
    except TypeError as err:
        raise get_type_error("cart_direction", cart_direction, "ndarray") from err
    cart_displacement = ref_structure.positions * 0
    try:
        cart_displacement[atom_index] = cart_direction
    except ValueError as err:
        raise get_shape_error("cart_direction", cart_direction, "(3,)") from err
    except IndexError as err:
        raise IndexError(f"invalid atom_index: {atom_index}") from err

    return get_displaced_positions(ref_structure, cart_displacement, amplitudes)


# pylint: disable=too-many-arguments,too-many-positional-arguments
def write_ast_displaced_structures(
    ref_structure: ReferenceStructure,
    atom_index: int,
    cart_direction: NDArray[np.float64],
    amplitudes: NDArray[np.float64],
    filepaths: str | Path | list[str] | list[Path],
    file_format: str,
    overwrite: bool = False,
) -> None:
    """Write displaced structures with a single atom displaced along a direction.

    Parameters
    ----------
    ref_structure
        Reference structure containing N atoms.
    atom_index
    cart_direction
        (Å) Array with shape (3,). The magnitude of the direction vector is ignored,
        only the direction is used.
    amplitudes
        (Å) Array with shape (M,).
    filepaths
    file_format
        Supports ``"poscar"`` (see :ref:`Supported formats`).
    overwrite
        Overwrite the file if it exists.
    """
    filepaths = pathify_as_list(filepaths)
    position_list = get_ast_displaced_positions(
        ref_structure, atom_index, cart_direction, amplitudes
    )
    verify_list_len("filepaths", filepaths, len(position_list))

    for position, filepath in zip(position_list, filepaths):
        generic_io.write_structure(
            ref_structure.lattice,
            ref_structure.atomic_numbers,
            position,
            filepath,
            file_format,
            overwrite,
        )



"""Functions for interacting with pymatgen."""

import numpy as np
from numpy.typing import NDArray

from ramannoodle.exceptions import (
    get_pymatgen_missing_error,
    UserError,
    verify_list_len,
    verify_ndarray_shape,
    get_type_error,
    IncompatibleStructureException,
)
from ramannoodle.dynamics._trajectory import Trajectory
from ramannoodle.structure._reference import ReferenceStructure

try:
    from ramannoodle.dataset.torch._dataset import PolarizabilityDataset
except UserError:
    pass

try:
    import pymatgen.core
    import pymatgen.core.trajectory
except ImportError as exc:
    raise get_pymatgen_missing_error() from exc


def _get_lattice(pymatgen_structure: pymatgen.core.Structure) -> NDArray[np.float64]:
    """Get lattice from a pymatgen Structure.

    Parameters
    ----------
    pymatgen_structure

    Returns
    -------
    :
        (Å) Array with shape (3,3).

    """
    try:
        return pymatgen_structure.lattice.matrix
    except AttributeError as exc:
        raise get_type_error(
            "pymatgen_structure", pymatgen_structure, "pymatgen.core.Structure"
        ) from exc


def get_positions(pymatgen_structure: pymatgen.core.Structure) -> NDArray[np.float64]:
    """Read fractional positions from a pymatgen Structure.

    Parameters
    ----------
    pymatgen_structure

    Returns
    -------
    :
        (fractional) Array with shape (N,3) where N is the number of atoms.

    """
    try:
        return pymatgen_structure.frac_coords
    except AttributeError as exc:
        raise get_type_error(
            "pymatgen_structure", pymatgen_structure, "pymatgen.core.Structure"
        ) from exc


def _get_atomic_numbers(pymatgen_structure: pymatgen.core.Structure) -> list[int]:
    """Get atomic numbers from a pymatgen Structure.

    Parameters
    ----------
    pymatgen_structure

    Returns
    -------
    :
        List of length N where N is the number of atoms.

    """
    try:
        return list(pymatgen_structure.atomic_numbers)
    except AttributeError as exc:
        raise get_type_error(
            "pymatgen_structure", pymatgen_structure, "pymatgen.core.Structure"
        ) from exc


def get_structure(
    pymatgen_structure: pymatgen.core.Structure,
) -> tuple[NDArray[np.float64], list[int], NDArray[np.float64]]:
    """Get lattice, positions, and atomic numbers from a pymatgen Structure.

    Parameters
    ----------
    pymatgen_structure

    Returns
    -------
    :
        0.  lattice -- (Å) Array with shape (3,3).
        1.  atomic numbers -- List of length N where N is the number of atoms.
        2.  positions -- (fractional) Array with shape (N,3) where N is the number of
            atoms.
    """
    return (
        _get_lattice(pymatgen_structure),
        _get_atomic_numbers(pymatgen_structure),
        get_positions(pymatgen_structure),
    )


def construct_polarizability_dataset(
    pymatgen_structures: list[pymatgen.core.Structure],
    polarizabilities: NDArray[np.float64],
) -> "PolarizabilityDataset":
    """Create a PolarizabilityDataset from of pymatgen Structures and polarizabilities.

    Parameters
    ----------
    pymatgen_structures
        List of length M.
    polarizabilities
        Array with shape (M,3,3).

    """
    verify_list_len("pymatgen_structures", pymatgen_structures, None)
    verify_ndarray_shape(
        "polarizabilities", polarizabilities, (len(pymatgen_structures), 3, 3)
    )
    lattice, atomic_numbers, _ = get_structure(pymatgen_structures[0])
    positions = np.zeros((len(pymatgen_structures), len(atomic_numbers), 3))
    for i, pymatgen_structure in enumerate(pymatgen_structures):
        if not np.allclose(_get_lattice(pymatgen_structure), lattice, atol=1e-5):
            raise IncompatibleStructureException(
                f"incompatible lattice: pymatgen_structures[{i}]"
            )
        if _get_atomic_numbers(pymatgen_structure) != atomic_numbers:
            raise IncompatibleStructureException(
                f"incompatible atomic numbers: pymatgen_structures[{i}]"
            )
        positions[i] = get_positions(pymatgen_structure)
    return PolarizabilityDataset(lattice, atomic_numbers, positions, polarizabilities)


def construct_ref_structure(
    pymatgen_structure: pymatgen.core.Structure,
) -> ReferenceStructure:
    """Create a ReferenceStructure from a pymatgen Structure.

    Parameters
    ----------
    pymatgen_structure

    """
    return ReferenceStructure(
        _get_atomic_numbers(pymatgen_structure),
        _get_lattice(pymatgen_structure),
        get_positions(pymatgen_structure),
    )


def construct_trajectory(
    pymatgen_trajectory: pymatgen.core.trajectory.Trajectory,
    timestep: float,
) -> Trajectory:
    """Create a Trajectory from a pymatgen Trajectory.

    Parameters
    ----------
    pymatgen_trajectory
    timestep
        (fs)

    """
    try:
        if not pymatgen_trajectory.constant_lattice:
            raise ValueError("pymatgen_trajectory must have a constant lattice")
    except AttributeError as exc:
        raise get_type_error(
            "pymatgen_trajectory",
            pymatgen_trajectory,
            "pymatgen.core.trajectory.Trajectory",
        ) from exc
    return Trajectory(pymatgen_trajectory.coords, timestep)



"""Utility functions for PyTorch models."""

from typing import Generator

import numpy as np
from numpy.typing import NDArray

from ramannoodle.exceptions import (
    get_torch_missing_error,
    verify_ndarray_shape,
)

try:
    import torch
    from torch import Tensor
    from torch_geometric.nn.models.dimenet import triplets as dimenet_triplets
except ModuleNotFoundError as exc:
    raise get_torch_missing_error() from exc

# pylint complains about torch.norm
# pylint: disable=not-callable


def get_rotations(targets: Tensor) -> Tensor:
    """Get rotation matrices from (1,0,0) to target vectors.

    Parameters
    ----------
    targets
        Tensor with size [S,3]. Vectors do not need to be normalized.

    Returns
    -------
    :
        Tensor with size [S,3,3].
    """
    reference = torch.zeros(targets.size())
    reference[:, 0] = 1

    a = reference / torch.linalg.norm(reference, dim=1).view(-1, 1)
    b = targets / torch.linalg.norm(targets, dim=1).view(-1, 1)

    v = torch.linalg.cross(a, b)  # This will be (0,0,0) if a == b
    c = torch.linalg.vecdot(a, b)
    s = torch.linalg.norm(v, dim=1)  # This will be zero if a == b

    k_matrix = torch.zeros((len(v), 3, 3))
    k_matrix[:, 0, 1] = -v[:, 2]
    k_matrix[:, 0, 2] = v[:, 1]
    k_matrix[:, 1, 0] = v[:, 2]
    k_matrix[:, 1, 2] = -v[:, 0]
    k_matrix[:, 2, 0] = -v[:, 1]
    k_matrix[:, 2, 1] = v[:, 0]

    rotations = torch.zeros((len(v), 3, 3))
    rotations[:] = torch.eye(3)  # Rotation starts as identity
    rotations += k_matrix
    a1 = k_matrix.matmul(k_matrix)
    b1 = (1 - c) / (s**2)
    rotations += a1 * b1[:, None, None]

    # a==b implies s==0, which implies rotation should be identity/
    rotations[s == 0] = torch.eye(3)

    return rotations


def get_graph_info(
    cart_displacement: Tensor,
    edge_indexes: Tensor,
    cart_distance_matrix: Tensor,
    num_atoms: int,
) -> tuple[Tensor, Tensor, Tensor]:
    """Get information on graph.

    :meta: private
    """
    cart_unit_vectors = cart_displacement[
        edge_indexes[0], edge_indexes[1], edge_indexes[2]
    ]  # python 3.10 complains if we use the unpacking operator (*)
    cart_unit_vectors /= torch.linalg.norm(cart_unit_vectors, dim=-1)[:, None]
    distances = cart_distance_matrix[
        edge_indexes[0], edge_indexes[1], edge_indexes[2]
    ].view(-1, 1)
    # ref_distances = self._ref_distances.repeat((num_samples, 1))
    # distances -= ref_distances
    # Disconnect all S graphs
    edge_indexes[1] += edge_indexes[0] * num_atoms
    edge_indexes[2] += edge_indexes[0] * num_atoms
    return edge_indexes, cart_unit_vectors, distances


def _radius_graph_pbc(
    lattice: Tensor, positions: Tensor, cutoff: float
) -> tuple[Tensor, Tensor, Tensor]:
    """Generate graph for structures while respecting periodic boundary conditions.

    Parameters
    ----------
    lattice
        (Å) Tensor with size [S,3,3] where S is the number of samples.
    positions
        (fractional) Tensor with size [S,N,3] where N is the number of atoms.
    cutoff
        Edge cutoff distance.

    Returns
    -------
    :
        0.  edge indexes -- Tensor of size [3,X] where X is the number of edges. This
            tensor defines S non-interconnected graphs making up a batch. The first row
            defines the graph index. The second and third rows define the actual edge
            indexes used by ``triplet``.
        1.  cartesian unit vectors -- (Å) Tensor with size [X,3].
        2.  distances -- (Å) Tensor with size [X,1].

    """
    num_samples = lattice.size(0)
    num_atoms = positions.size(1)

    # Compute pairwise distance matrix.
    displacement = positions.unsqueeze(1) - positions.unsqueeze(2)
    displacement = torch.where(
        displacement % 1 > 0.5, displacement % 1 - 1, displacement % 1
    )
    expanded_lattice = lattice[:, None, :, :].expand(-1, displacement.size(1), -1, -1)
    cart_displacement = displacement.matmul(expanded_lattice)
    cart_distance_matrix = torch.sqrt(torch.sum(cart_displacement**2, dim=-1))

    # Compute adjacency matrix
    adjacency_matrix = cart_distance_matrix <= cutoff
    not_self_loop = ~torch.eye(adjacency_matrix.size(-1), dtype=torch.bool).expand(
        num_samples, -1, -1
    )
    adjacency_matrix = torch.logical_and(adjacency_matrix, not_self_loop)

    edge_indexes = torch.nonzero(adjacency_matrix).T

    return get_graph_info(
        cart_displacement, edge_indexes, cart_distance_matrix, num_atoms
    )


class BatchTriplets:
    """Graph triplets.

    Parameters
    ----------
    num_nodes
    ref_edge_indexes
    """

    def __init__(self, num_nodes: int, ref_edge_indexes: Tensor):
        self._num_nodes = num_nodes
        self._num_edges = len(ref_edge_indexes[1])
        with torch.device("cpu"):
            self._ref_triplets = dimenet_triplets(
                edge_index=ref_edge_indexes[[1, 2]].to("cpu"),
                num_nodes=self._num_nodes,
            )
        self._ref_triplets = tuple(
            t.to(torch.get_default_device()) for t in self._ref_triplets
        )
        assert len(self._ref_triplets) == 7  # type hint
        self._num_triplets = self._ref_triplets[2].size(0)

        self._cached_batch_size = 1
        self._cached_triplets = self._ref_triplets

    @property
    def cached_batch_size(self) -> int:
        """Get batch size of current cache."""
        return self._cached_batch_size

    def get_triplets(
        self, batch_size: int
    ) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
        """Get triplets for a given batch size.

        Returns
        -------
        :
            0.  i -- Node 1 of edge pairs, a tensor with size [E,].
            #.  j -- Node 2 of edge pairs, a tensor with size [E,].
            #.  index_i -- Node 1 of edge triplets, a tensor with size [T,] where T is
                the number of triplets.
            #.  index_j -- Node 2 of edge triplets, a tensor with size [T,].
            #.  index_k -- Node 3 of edge triplets, a tensor with size [T,].
            #.  index_ji -- Index of (j,i) corresponding to (index_j,index_i), a tensor
                with size [T,].
            #.  index_kj -- Index of (k,j) corresponding to (index_k,index_j), a tensor
                with size [T,].

        """
        if batch_size != self._cached_batch_size:
            i, j, index_i, index_j, index_k, index_ji, index_kj = tuple(
                t.repeat(batch_size) for t in self._ref_triplets
            )

            batch_indexes = torch.tensor([range(batch_size)]).repeat_interleave(
                self._num_edges, dim=1
            )[0]
            for index in [i, j]:
                index += batch_indexes * self._num_nodes

            batch_indexes = torch.tensor([range(batch_size)]).repeat_interleave(
                self._num_triplets, dim=1
            )[0]
            for index in [index_i, index_j, index_k]:
                index += batch_indexes * self._num_nodes
            for index in [index_ji, index_kj]:
                index += batch_indexes * self._num_edges

            self._cached_triplets = (
                i,
                j,
                index_i,
                index_j,
                index_k,
                index_ji,
                index_kj,
            )
            self._cached_batch_size = batch_size

        return self._cached_triplets


def batch_positions(
    positions: NDArray[np.float64], batch_size: int
) -> Generator[NDArray[np.float64], None, None]:
    """Split positions into batches.

    Parameters
    ----------
    positions
        (fractional) Array with shape (S,N,3) where S is the number of samples
        and N is the number of atoms.
    batch_size
        Split positions into batches of size ``batch_size``.

    Yields
    ------
    :
        Array with shape (batch_size,N,3).

    """
    verify_ndarray_shape("positions", positions, (None, None, 3))
    for batch_index in range(0, positions.shape[0], batch_size):
        yield positions[batch_index : min(batch_index + batch_size, positions.shape[0])]



"""Generic IO functions.

Generic IO functions are somewhat inflexible but are necessary for certain
functionality. Users are strongly encouraged to use IO functions contained in the
code-specific subpackages. For example, IO for VASP POSCAR and OUTCAR files can be
accomplished using :mod:`ramannoodle.io.vasp.poscar` and
:mod:`ramannoodle.io.vasp.outcar` respectively.

"""

from pathlib import Path

import numpy as np
from numpy.typing import NDArray

from ramannoodle.dynamics._phonon import Phonons
from ramannoodle.dynamics._trajectory import Trajectory
from ramannoodle.structure._reference import ReferenceStructure
from ramannoodle.exceptions import UserError, get_torch_missing_error
import ramannoodle.io.vasp as vasp_io

TORCH_PRESENT = True
try:
    from ramannoodle.dataset.torch._dataset import PolarizabilityDataset
except UserError:
    TORCH_PRESENT = False

# These  map between file formats and appropriate IO functions.
_PHONON_READERS = {
    "outcar": vasp_io.outcar.read_phonons,
    "vasprun.xml": vasp_io.vasprun.read_phonons,
}
_TRAJECTORY_READERS = {
    "outcar": vasp_io.outcar.read_trajectory,
    "vasprun.xml": vasp_io.vasprun.read_trajectory,
}
_POSITION_AND_POLARIZABILITY_READERS = {
    "outcar": vasp_io.outcar.read_positions_and_polarizability,
    "vasprun.xml": vasp_io.vasprun.read_positions_and_polarizability,
}
_STRUCTURE_AND_POLARIZABILITY_READERS = {
    "outcar": vasp_io.outcar.read_structure_and_polarizability,
    "vasprun.xml": vasp_io.vasprun.read_structure_and_polarizability,
}
_POLARIZABILITY_DATASET_READERS = {
    "outcar": vasp_io.outcar.read_polarizability_dataset,
    "vasprun.xml": vasp_io.vasprun.read_polarizability_dataset,
}
_POSITION_READERS = {
    "poscar": vasp_io.poscar.read_positions,
    "outcar": vasp_io.outcar.read_positions,
    "xdatcar": vasp_io.poscar.read_positions,
    "vasprun.xml": vasp_io.vasprun.read_positions,
}
_REFERENCE_STRUCTURE_READERS = {
    "poscar": vasp_io.poscar.read_ref_structure,
    "outcar": vasp_io.outcar.read_ref_structure,
    "xdatcar": vasp_io.poscar.read_ref_structure,
    "vasprun.xml": vasp_io.vasprun.read_ref_structure,
}
_STRUCTURE_WRITERS = {
    "poscar": vasp_io.poscar.write_structure,
    "xdatcar": vasp_io.poscar.write_structure,
}
_TRAJECTORY_WRITERS = {"xdatcar": vasp_io.xdatcar.write_trajectory}


def read_phonons(filepath: str | Path, file_format: str) -> Phonons:
    """Read phonons from a file.

    Parameters
    ----------
    filepath
    file_format
        Supports ``"outcar"``, ``"vasprun.xml"`` (see :ref:`Supported formats`).

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    try:
        return _PHONON_READERS[file_format](filepath)
    except KeyError as exc:
        raise ValueError(f"unsupported format: {file_format}") from exc


def read_trajectory(filepath: str | Path, file_format: str) -> Trajectory:
    """Read molecular dynamics trajectory from a file.

    Parameters
    ----------
    filepath
    file_format
        Supports ``"outcar"``, ``"vasprun.xml"``, (see :ref:`Supported formats`).
        Use :func:`.vasp.xdatcar.read_trajectory` to read a trajectory from an
        XDATCAR.

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    try:
        return _TRAJECTORY_READERS[file_format](filepath)
    except KeyError as exc:
        if file_format == "xdatcar":
            raise ValueError(
                "generic.read_trajectory does not support xdatcar."
            ) from exc
        raise ValueError(f"unsupported format: {file_format}") from exc


def read_positions_and_polarizability(
    filepath: str | Path,
    file_format: str,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Read fractional positions and polarizability from a file.

    Parameters
    ----------
    filepath
    file_format
        Supports ``"outcar"``, ``"vasprun.xml"`` (see :ref:`Supported formats`).

    Returns
    -------
    :
        0.  positions -- (fractional) Array with shape (N,3) where N is the number of
            atoms.
        #.  polarizability -- Array with shape (3,3).

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    try:
        return _POSITION_AND_POLARIZABILITY_READERS[file_format](filepath)
    except KeyError as exc:
        raise ValueError(f"unsupported format: {file_format}") from exc


def read_structure_and_polarizability(
    filepath: str | Path,
    file_format: str,
) -> tuple[NDArray[np.float64], list[int], NDArray[np.float64], NDArray[np.float64]]:
    """Read lattice, atomic numbers, fractional positions, polarizability from a file.

    Parameters
    ----------
    filepath
    file_format
        Supports ``"outcar"``, ``"vasprun.xml"`` (see :ref:`Supported formats`)

    Returns
    -------
    :
        0.  lattice -- (Å) Array with shape (3,3).
        #.  atomic_numbers -- List of length N where N is the number of atoms.
        #.  positions -- (fractional) Array with shape (N,3) where N is the number of
            atoms.
        #.  polarizability -- Array with shape (3,3).

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    try:
        return _STRUCTURE_AND_POLARIZABILITY_READERS[file_format](filepath)
    except KeyError as exc:
        raise ValueError(f"unsupported format: {file_format}") from exc


def read_polarizability_dataset(
    filepaths: str | Path | list[str] | list[Path],
    file_format: str,
) -> "PolarizabilityDataset":
    """Read polarizability dataset from files.

    Parameters
    ----------
    filepaths
    file_format
        Supports ``"outcar"``, ``"vasprun.xml"`` (see :ref:`Supported formats`)

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    IncompatibleFileException
        File is incompatible with the dataset.
    """
    if not TORCH_PRESENT:
        raise get_torch_missing_error()
    try:
        return _POLARIZABILITY_DATASET_READERS[file_format](filepaths)
    except KeyError as exc:
        raise ValueError(f"unsupported format: {file_format}") from exc


def read_positions(
    filepath: str | Path,
    file_format: str,
) -> NDArray[np.float64]:
    """Read fractional positions from a file.

    Parameters
    ----------
    filepath
    file_format
        Supports ``"outcar"``, ``"poscar"``, ``"xdatcar"``, ``"vasprun.xml"``  (see
        :ref:`Supported formats`).

    Returns
    -------
    :
        (fractional) Array with shape (N,3) where N is the number of atoms.

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.

    """
    try:
        return _POSITION_READERS[file_format](filepath)
    except KeyError as exc:
        raise ValueError(f"unsupported format: {file_format}") from exc


def read_ref_structure(filepath: str | Path, file_format: str) -> ReferenceStructure:
    """Read reference structure from a file.

    Parameters
    ----------
    filepath
    file_format
        Supports ``"outcar"``, ``"poscar"``, ``"xdatcar"``, ``"vasprun.xml"`` (see
        :ref:`Supported formats`).

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    SymmetryException
        Structural symmetry determination failed.
    """
    try:
        return _REFERENCE_STRUCTURE_READERS[file_format](filepath)
    except KeyError as exc:
        raise ValueError(f"unsupported format: {file_format}") from exc


def write_structure(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    lattice: NDArray[np.float64],
    atomic_numbers: list[int],
    positions: NDArray[np.float64],
    filepath: str | Path,
    file_format: str,
    overwrite: bool = False,
) -> None:
    """Write structure to file.

    Parameters
    ----------
    lattice
        (Å) Array with shape (3,3).
    atomic_numbers
        List of length N where N is the number of atoms.
    positions
        (fractional) Array with shape (N,3).
    filepath
    file_format
        Supports ``"poscar"`` (see :ref:`Supported formats`).
    overwrite
        Overwrite the file if it exists.
    label
        POSCAR label (first line).

    Raises
    ------
    FileExistsError
        File exists and ``overwrite == False``.
    """
    try:
        _STRUCTURE_WRITERS[file_format](
            lattice=lattice,
            atomic_numbers=atomic_numbers,
            positions=positions,
            filepath=filepath,
            overwrite=overwrite,
        )
    except KeyError as exc:
        raise ValueError(f"unsupported format: {file_format}") from exc


# pylint: disable=too-many-arguments,too-many-positional-arguments
def write_trajectory(
    lattice: NDArray[np.float64],
    atomic_numbers: list[int],
    positions_ts: NDArray[np.float64],
    filepath: str | Path,
    file_format: str,
    overwrite: bool = False,
) -> None:
    """Write trajectory to file.

    Parameters
    ----------
    lattice
        (Å) Array with shape (3,3).
    atomic_numbers
        List of length N where N is the number of atoms.
    positions_ts
        (fractional) Array with shape (S,N,3) where S is the number of
        configurations.
    filepath
    file_format
        Supports ``"xdatcar"`` (see :ref:`Supported formats`).
    overwrite
        Overwrite the file if it exists.

    Raises
    ------
    FileExistsError
        File exists and ``overwrite == False``.
    """
    try:
        _TRAJECTORY_WRITERS[file_format](
            lattice=lattice,
            atomic_numbers=atomic_numbers,
            positions_ts=positions_ts,
            filepath=filepath,
            overwrite=overwrite,
        )
    except KeyError as exc:
        raise ValueError(f"unsupported format: {file_format}") from exc



"""Utility functions for structures."""

import numpy as np
from numpy.typing import NDArray

from ramannoodle.exceptions import (
    get_type_error,
    verify_positions,
    get_shape_error,
)


def apply_pbc(positions: NDArray[np.float64]) -> NDArray[np.float64]:
    """Return fractional positions such that all coordinates are between 0 and 1.

    Parameters
    ----------
    positions
        (fractional) Array with shape (N,3) where N is the number of atoms.

    Returns
    -------
    :
        (fractional) Array with shape (N,3).
    """
    try:
        return positions - positions // 1
    except TypeError as exc:
        raise get_type_error("positions", positions, "ndarray") from exc


def apply_pbc_displacement(displacement: NDArray[np.float64]) -> NDArray[np.float64]:
    """Return fractional displacement such as all coordinates are between -0.5 and 0.5.

    Parameters
    ----------
    displacement
        (fractional) Array with shape (N,3) where N is the number of atoms.

    Returns
    -------
    :
        (fractional) Array with shape (N,3).
    """
    try:
        return np.where(displacement % 1 > 0.5, displacement % 1 - 1, displacement % 1)
    except TypeError as exc:
        raise get_type_error("displacement", displacement, "ndarray") from exc


def displace_positions(
    positions: NDArray[np.float64],
    displacement: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Displace positions, respecting periodic boundary conditions.

    Parameters
    ----------
    positions
        (fractional) Array with shape (N,3) where N is the number of atoms.
    displacement
        (fractional) Array with shape (N,3).

    Returns
    -------
    :
        (fractional) Array with shape (N,3).
    """
    positions = apply_pbc(positions)
    displacement = apply_pbc_displacement(displacement)

    return apply_pbc(positions + displacement)


def transform_positions(
    positions: NDArray[np.float64],
    rotation: NDArray[np.float64],
    translation: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Transform positions.

    Respects periodic boundary conditions.

    Parameters
    ----------
    positions
        (fractional) Array with shape (N,3) where N is the number of atoms
    rotation
        Array with shape (3,3).
    translation
        (fractional) Array with shape (3,).

    Returns
    -------
    :
        (fractional) Array with shape (N,3).
    """
    verify_positions("positions", positions)
    positions = apply_pbc(positions)
    try:
        rotated = positions @ rotation
    except TypeError as exc:
        raise get_type_error("rotation", rotation, "ndarray") from exc
    except ValueError as exc:
        raise get_shape_error("rotation", rotation, "(3,3)") from exc
    rotated = apply_pbc(rotated)
    return displace_positions(rotated, translation)


def calc_displacement(
    positions_1: NDArray[np.float64],
    positions_2: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Calculate minimum displacement between two fractional positions.

    Displacement is from ``positions_1`` to ``positions_2``. Respects periodic boundary
    conditions.

    Parameters
    ----------
    positions_1
        (fractional) Array with shape (N,3) where N is the number of atoms.
    positions_2
        (fractional) Array with shape (N,3).

    Returns
    -------
    :
        (fractional) Array with shape (N,3).

    """
    positions_1 = apply_pbc(positions_1)
    positions_2 = apply_pbc(positions_2)

    return apply_pbc_displacement(positions_2 - positions_1)



"""Polarizability model based on interpolation around degrees of freedom."""

# This is not ideal, but is required for Python 3.10 support.
# In future versions, we can use "from typing import Self"
from __future__ import annotations

from pathlib import Path
import copy
from typing import Iterable
from warnings import warn
import itertools

import numpy as np
from numpy.typing import NDArray, ArrayLike
from scipy.interpolate import make_interp_spline, BSpline

from ramannoodle.constants import ANSICOLORS
from ramannoodle.abstract import PolarizabilityModel
from ramannoodle.structure.utils import calc_displacement
from ramannoodle.structure._symmetry_utils import (
    is_orthogonal_to_all,
    is_collinear_with_all,
)
from ramannoodle.structure._reference import ReferenceStructure
from ramannoodle.exceptions import (
    InvalidDOFException,
    get_type_error,
    get_shape_error,
    verify_ndarray_shape,
    DOFWarning,
    UserError,
    verify_list_len,
    get_pymatgen_missing_error,
)
import ramannoodle.io.generic as generic_io
from ramannoodle.io._utils import pathify_as_list

PYMATGEN_PRESENT = True
try:
    import ramannoodle.io.pymatgen as pymatgen_io
    import pymatgen.core
except (UserError, ImportError):
    PYMATGEN_PRESENT = False


def get_amplitude(
    cart_basis_vector: NDArray[np.float64],
    cart_displacement: NDArray[np.float64],
) -> float:
    """Get amplitude of a displacement.

    Returns
    -------
    :
        (Å)
    """
    return float(np.dot(cart_basis_vector.flatten(), cart_displacement.flatten()))


def find_duplicates(vectors: Iterable[ArrayLike]) -> NDArray | None:
    """Return duplicate vector in a list or None if no duplicates found.

    :meta private:
    """
    try:
        combinations = itertools.combinations(vectors, 2)
    except TypeError as exc:
        raise get_type_error("vectors", vectors, "Iterable") from exc
    try:
        for vector_1, vector_2 in combinations:
            if np.allclose(vector_1, vector_2):
                return np.array(vector_1)
        return None
    except TypeError as exc:
        raise TypeError("elements of vectors are not array_like") from exc


class InterpolationModel(PolarizabilityModel):
    """Polarizability model based on interpolation around degrees of freedom (DOFs).

    One is free to specify the interpolation order as well as the precise
    the DOFs, so long as they are orthogonal. For example, one can
    employ first-order (linear) interpolation around phonon displacements to calculate
    a conventional Raman spectrum. One can achieve identical results -- often with fewer
    calculations -- by using first-order interpolations around atomic displacements.

    This model's key assumption is that each degree of freedom in a system modulates
    the polarizability **independently**.

    Parameters
    ----------
    ref_structure
        Reference structure on which to base the model.
    ref_polarizability
        Array with shape (3,3) with polarizability of the reference structure.
    is_dummy_model

    """

    def __init__(
        self,
        ref_structure: ReferenceStructure,
        ref_polarizability: NDArray[np.float64],
        is_dummy_model: bool = False,
    ) -> None:
        if is_dummy_model:
            ref_polarizability = np.zeros((3, 3))
        verify_ndarray_shape("ref_polarizability", ref_polarizability, (3, 3))

        self._ref_structure = ref_structure
        self._ref_polarizability = ref_polarizability
        self._is_dummy_model = is_dummy_model
        self._cart_basis_vectors: list[NDArray[np.float64]] = []
        self._interpolations: list[BSpline] = []
        self._mask: NDArray[np.bool] = np.array([], dtype="bool")

    @property
    def ref_structure(self) -> ReferenceStructure:
        """Get (a copy of) reference structure."""
        return copy.deepcopy(self._ref_structure)

    @property
    def ref_polarizability(self) -> NDArray[np.float64]:
        """Get (a copy of) reference polarizability.

        Returns
        -------
        :
            Array with shape (3,3).
        """
        return self._ref_polarizability.copy()

    @property
    def is_dummy_model(self) -> bool:
        """Get whether model is a dummy model."""
        return self._is_dummy_model

    @property
    def cart_basis_vectors(self) -> list[NDArray[np.float64]]:
        """Get (a copy of) cartesian basis vectors.

        Returns
        -------
        :
            (Å) List of length J containing arrays with shape (N,3) where J is the
            number of specified degrees of freedom and N is the number of atoms.

        """
        return copy.deepcopy(self._cart_basis_vectors)

    @property
    def interpolations(self) -> list[BSpline]:
        """Get (a copy of) interpolations.

        Returns
        -------
        :
            List of length J where J is the number of specified degrees of freedom.
        """
        return copy.deepcopy(self._interpolations)

    @property
    def mask(self) -> NDArray[np.bool]:
        """Get (a copy of) mask.

        Returns
        -------
        :
            Array with shape (J,) where J is the number of specified degrees of
            freedom.
        """
        return self._mask.copy()

    @mask.setter
    def mask(self, value: NDArray[np.bool]) -> None:
        """Set mask.

        ..warning:: To avoid unintentional use of masked models, we discourage masking
                    in-place. Instead, consider using :meth:`get_masked_model`.

        Parameters
        ----------
        mask
            Array with shape (N,) where N is the number of specified degrees
            of freedom (DOFs). If an element is False, its corresponding DOF will be
            "masked" and excluded from polarizability calculations.
        """
        verify_ndarray_shape("mask", value, self._mask.shape)
        self._mask = value

    def calc_polarizabilities(
        self, positions_batch: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Return estimated polarizabilities for a batch of fractional positions.

        Parameters
        ----------
        positions_batch
            (fractional) Array with shape (S,N,3) where S is the number of samples
            and N is the number of atoms.

        Returns
        -------
        :
            Array with shape (S,3,3).

        Raises
        ------
        UserError
            Model is a dummy model.

        """
        try:
            delta_polarizabilities: NDArray[np.float64] = np.zeros(
                (positions_batch.shape[0], 3, 3)
            )
            cart_displacements = self._ref_structure.get_cart_displacement(
                calc_displacement(self._ref_structure.positions, positions_batch)
            )
            cart_displacements = cart_displacements.reshape(
                cart_displacements.shape[0],
                cart_displacements.shape[1] * cart_displacements.shape[2],
            )
        except (AttributeError, TypeError) as exc:
            raise get_type_error("positions", positions_batch, "ndarray") from exc
        except (ValueError, IndexError) as exc:
            raise get_shape_error(
                "positions", positions_batch, f"(_,{self._ref_structure.num_atoms},3)"
            ) from exc

        try:

            for basis_vector, interpolation, mask in zip(
                self._cart_basis_vectors,
                self._interpolations,
                self._mask,
                strict=True,
            ):
                amplitudes = np.einsum(
                    "i,ji", basis_vector.flatten(), cart_displacements
                )
                delta_polarizabilities += (1 - mask) * np.array(
                    interpolation(amplitudes), dtype="float64"
                )
        except ValueError as err:
            if self._is_dummy_model:
                raise UserError(
                    "dummy model cannot calculate polarizabilities"
                ) from err
            raise err

        return delta_polarizabilities + self._ref_polarizability

    def _get_dof(  # pylint: disable=too-many-locals
        self,
        parent_displacement: NDArray[np.float64],
        amplitudes: NDArray[np.float64],
        polarizabilities: NDArray[np.float64],
        include_ref_polarizability: bool,
    ) -> tuple[
        list[NDArray[np.float64]], list[list[float]], list[list[NDArray[np.float64]]]
    ]:
        """Calculate and return basis vectors and interpolation points for DOF(s).

        This method will check the displacement to make sure it's valid, but will not
        check the amplitudes. Amplitude checking is performed in
        ``_construct_and_add_interpolations``, as this method has access to the
        interpolation order.

        Parameters
        ----------
        parent_displacement
            Displacement of the parent DOF.
        amplitudes
            Amplitudes (of the parent DOF).
        polarizabilities
            Polarizabilities (of the parent DOF).

        Returns
        -------
        :
            0.  basis vectors -- (Å) List of length J containing arrays with shape (N,3)
                where J is the number of degrees of freedom and N is the number of
                atoms.
            #.  interpolation_xs
            #.  interpolation_ys
        """
        # Check that the parent displacement is orthogonal to existing basis vectors
        parent_cart_basis_vector = self._ref_structure.get_cart_displacement(
            parent_displacement
        )
        result = is_orthogonal_to_all(
            parent_cart_basis_vector, self._cart_basis_vectors
        )
        if result != -1:
            raise InvalidDOFException(
                f"new dof is not orthogonal with existing dof (index={result})"
            )

        displacements_and_transformations = (
            self._ref_structure.get_equivalent_displacements(parent_displacement)
        )

        basis_vectors: list[NDArray[np.float64]] = []
        interpolation_xs: list[list[float]] = []
        interpolation_ys: list[list[NDArray[np.float64]]] = []
        for dof_dictionary in displacements_and_transformations:
            child_displacement = dof_dictionary["displacements"][0]

            interpolation_x: list[float] = []
            interpolation_y: list[NDArray[np.float64]] = []
            if include_ref_polarizability:
                interpolation_x.append(0.0)
                interpolation_y.append(np.zeros((3, 3)))

            for collinear_displacement, transformation in zip(
                dof_dictionary["displacements"], dof_dictionary["transformations"]
            ):
                _index = np.unravel_index(
                    np.argmax(np.abs(child_displacement)), child_displacement.shape
                )
                multiplier = child_displacement[_index] / collinear_displacement[_index]

                for amplitude, polarizability in zip(amplitudes, polarizabilities):
                    interpolation_x.append(multiplier * amplitude)
                    rotation = transformation[0]
                    delta_polarizability = polarizability - self._ref_polarizability

                    interpolation_y.append(
                        (np.linalg.inv(rotation) @ delta_polarizability @ rotation)
                    )

            child_cart_basis_vector = self._ref_structure.get_cart_displacement(
                child_displacement
            )
            child_cart_basis_vector /= np.linalg.norm(child_cart_basis_vector)

            basis_vectors.append(child_cart_basis_vector)
            interpolation_xs.append(interpolation_x)
            interpolation_ys.append(interpolation_y)

        return (basis_vectors, interpolation_xs, interpolation_ys)

    def _construct_and_add_interpolations(
        self,
        basis_vectors_to_add: list[NDArray[np.float64]],
        interpolation_xs: list[list[float]],
        interpolation_ys: list[list[NDArray[np.float64]]],
        interpolation_order: int,
    ) -> None:
        """Construct interpolations and add them to the model.

        Raises
        ------
        InvalidDOFException
        """
        # make_interp_spline accepts k=0 (useless), so check here
        if not isinstance(interpolation_order, int):
            raise get_type_error("interpolation_order", interpolation_order, "int")
        if interpolation_order < 1:
            raise ValueError(f"invalid interpolation_order: {interpolation_order} < 1")

        interpolations_to_add: list[BSpline] = []
        for interpolation_x, interpolation_y in zip(interpolation_xs, interpolation_ys):

            # Duplicate amplitudes means too much data has been provided.
            duplicate = find_duplicates(interpolation_x)
            if duplicate is not None:
                raise InvalidDOFException(
                    f"due to symmetry, amplitude {duplicate} should not be specified"
                )

            if len(interpolation_x) <= interpolation_order:
                raise InvalidDOFException(
                    f"insufficient points ({len(interpolation_x)}) available for "
                    f"{interpolation_order}-order interpolation"
                )

            # Warn user if amplitudes don't span zero
            max_amplitude = np.max(interpolation_x)
            min_amplitude = np.min(interpolation_x)
            if np.allclose(max_amplitude, 0, atol=1e-3) or max_amplitude <= 0:
                warn(
                    "max amplitude <= 0, when usually it should be > 0",
                    DOFWarning,
                )
            if np.allclose(min_amplitude, 0, atol=1e-3) or min_amplitude >= 0:
                warn(
                    "min amplitude >= 0, when usually it should be < 0",
                    DOFWarning,
                )

            sort_indices = np.argsort(interpolation_x)
            interpolations_to_add.append(
                make_interp_spline(
                    x=np.array(interpolation_x)[sort_indices],
                    y=np.array(interpolation_y)[sort_indices],
                    k=interpolation_order,
                    bc_type=None,
                )
            )

        self._cart_basis_vectors += basis_vectors_to_add
        if not self._is_dummy_model:
            self._interpolations += interpolations_to_add
        # FALSE -> not masking, TRUE -> masking
        self._mask = np.append(self._mask, [False] * len(basis_vectors_to_add))

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def add_dof(
        self,
        cart_displacement: NDArray[np.float64],
        amplitudes: NDArray[np.float64],
        polarizabilities: NDArray[np.float64],
        interpolation_order: int,
        include_ref_polarizability: bool = True,
    ) -> None:
        """Add a degree of freedom (DOF).

        Specification of a DOF requires a displacement (directions the atoms move)
        alongside displacement amplitudes and corresponding known polarizabilities for
        each amplitude. Alongside the DOF specified, all DOFs related by the system's
        symmetry will be added as well. The interpolation order can be specified,
        though one must ensure that sufficient data is available.

        Parameters
        ----------
        cart_displacement
            (Å) Array with shape (N,3) where N is the number of atoms. The magnitude of
            the displacement is ignored, only the direction is used. Must be orthogonal
            to all previously added DOFs.
        amplitudes
            (Å) Array with shape (L,). Duplicate amplitudes, either those explicitly
            provided or those generated by structural symmetries, will raise
            :class:`.InvalidDOFException`.
        polarizabilities
            Array with shape (1,3,3) or (2,3,3) containing known
            polarizabilities for each amplitude. If dummy model, this parameter is
            ignored.
        interpolation_order
            Must be less than the number of total number of amplitudes after
            symmetry considerations.
        include_ref_polarizability
            Whether to include the references polarizability at 0.0 amplitude in the
            interpolation.

        Raises
        ------
        InvalidDOFException
            Provided degree of freedom was invalid.

        """
        try:
            displacement = self.ref_structure.get_frac_displacement(cart_displacement)

            parent_displacement = displacement / np.linalg.norm(displacement * 10.0)
        except TypeError as exc:
            raise TypeError("displacement is not an ndarray") from exc
        verify_ndarray_shape("amplitudes", amplitudes, (None,))
        if self._is_dummy_model:
            polarizabilities = np.zeros((len(amplitudes), 3, 3))
        verify_ndarray_shape(
            "polarizabilities", polarizabilities, (len(amplitudes), 3, 3)
        )

        # Get information needed for DOF - checks displacement.
        basis_vectors_to_add, interpolation_xs, interpolation_ys = self._get_dof(
            parent_displacement,
            amplitudes,
            polarizabilities,
            include_ref_polarizability,
        )

        # Then append the DOF - checks amplitudes (in form of interpolation_xs)
        self._construct_and_add_interpolations(
            basis_vectors_to_add,
            interpolation_xs,
            interpolation_ys,
            interpolation_order,
        )

    def add_dof_from_files(
        self,
        filepaths: str | Path | list[str] | list[Path],
        file_format: str,
        interpolation_order: int,
    ) -> None:
        """Add a degree of freedom (DOF) from file(s).

        Required displacements, amplitudes, and polarizabilities are automatically
        determined from provided files. Files should be chosen such that the
        resulting DOFs are valid under the same restrictions of :meth:`add_dof`.

        Parameters
        ----------
        filepaths
        file_format
            Supports ``"outcar"`` and ``"vasprun.xml"``. If dummy model, supports
            ``"poscar"`` and ``"xdatcar"`` as well (see :ref:`Supported formats`).



        Raises
        ------
        FileNotFoundError
            File not found.
        InvalidFileException
            Invalid file.
        InvalidDOFException
            DOF assembled from supplied files was invalid. See :meth:`add_dof` for
            restrictions.


        """
        # Checks displacements
        displacements, amplitudes, polarizabilities = self._read_dof(
            filepaths, file_format
        )

        # Checks amplitudes
        self.add_dof(
            self.ref_structure.get_cart_displacement(displacements[0]),
            amplitudes,
            polarizabilities,
            interpolation_order,
        )

    def _get_displacement_amplitudes_pymatgen(
        self,
        pymatgen_structures: "list[pymatgen.core.Structure]",
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Get displacement and amplitudes from pymatgen Structures.

        Parameters
        ----------
        pymatgen_structures
            List of length M.

        Returns
        -------
        :
            0.  displacements -- (fractional) Array with shape (M,N,3) where N is the
                number of atoms.
            #.  amplitudes -- Array with shape (M,2).

        Raises
        ------
        InvalidDOFException
            DOF assembled from supplied Structures was invalid. See :meth:`add_dof` for
            restrictions.
        """
        if not PYMATGEN_PRESENT:
            raise get_pymatgen_missing_error()

        # Calculate displacements and basis vector
        displacements = []
        for i, structure in enumerate(pymatgen_structures):
            positions = pymatgen_io.get_positions(structure)
            try:
                displacement = calc_displacement(
                    self._ref_structure.positions,
                    positions,
                )
            except ValueError as exc:
                invalid = f"pymatgen_structures[{i}]"
                raise InvalidDOFException(f"incompatible structure: {invalid}") from exc
            displacements.append(displacement)
        result = is_collinear_with_all(displacements[0], displacements)
        if result != -1:
            invalid = f"pymatgen_structures[{result}]"
            raise InvalidDOFException(f"displacement ({invalid}) is not collinear")
        cart_basis_vector = self._ref_structure.get_cart_displacement(displacements[0])
        cart_basis_vector /= np.linalg.norm(cart_basis_vector)

        # Calculate amplitudes
        amplitudes = []
        for displacement in displacements:
            cart_displacement = self._ref_structure.get_cart_displacement(displacement)
            amplitudes.append(get_amplitude(cart_basis_vector, cart_displacement))
        return (
            np.array(displacements),
            np.array(amplitudes),
        )

    def add_dof_from_pymatgen(
        self,
        pymatgen_structures: list[pymatgen.core.Structure],
        polarizabilities: NDArray[np.float64],
        interpolation_order: int,
    ) -> None:
        """
        Add a degree of freedom from a list of pymatgen Structures and polarizabilities.

        Parameters
        ----------
        pymatgen_structures
            List of length M.
        polarizabilities
            Array with shape (M,3,3).
        interpolation_order
            Must be less than the number of total number of amplitudes after
            symmetry considerations.

        Raises
        ------
        InvalidDOFException
            DOF assembled from supplied structures was invalid. See :meth:`add_dof` for
            restrictions.
        UserError
            pymatgen is not installed.

        """
        if not PYMATGEN_PRESENT:
            raise get_pymatgen_missing_error()
        verify_list_len("pymatgen_structures", pymatgen_structures, None)

        # Checks displacements
        displacements, amplitudes = self._get_displacement_amplitudes_pymatgen(
            pymatgen_structures
        )
        # Checks amplitudes
        self.add_dof(
            self.ref_structure.get_cart_displacement(displacements[0]),
            amplitudes,
            polarizabilities,
            interpolation_order,
        )

    def _read_dof(
        self, filepaths: str | Path | list[str] | list[Path], file_format: str
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        """Read displacements, amplitudes, and polarizabilities from file(s).

        This function does not change the state of the model.

        Parameters
        ----------
        filepaths:
            Supports ``"outcar"``. If dummy model, supports ``"outcar"``, | |
            ``"poscar"`` (see :ref:`Supported formats`).

        Returns
        -------
        :
            0.  displacements -- (fractional) Array with shape (J,N,3) where J is the
                total number of displacements.
            #.  amplitudes
            #.  polarizabilities

        Raises
        ------
        FileNotFoundError
            File not found.
        InvalidDOFException
            DOF assembled from supplied files was invalid (see get_dof)
        """
        displacements = []
        polarizabilities = []
        filepaths = pathify_as_list(filepaths)
        for filepath in filepaths:
            if not self._is_dummy_model:
                positions, polarizability = (
                    generic_io.read_positions_and_polarizability(filepath, file_format)
                )
            else:
                positions = generic_io.read_positions(filepath, file_format)
                polarizability = np.zeros((3, 3))

            try:
                displacement = calc_displacement(
                    self._ref_structure.positions,
                    positions,
                )
            except ValueError as exc:
                raise InvalidDOFException(f"incompatible outcar: {filepath}") from exc
            displacements.append(displacement)
            polarizabilities.append(polarizability)
        result = is_collinear_with_all(displacements[0], displacements)
        if result != -1:
            raise InvalidDOFException(
                f"displacement (file-index={result}) is not collinear"
            )
        cart_basis_vector = self._ref_structure.get_cart_displacement(displacements[0])
        cart_basis_vector /= np.linalg.norm(cart_basis_vector)

        # Calculate amplitudes
        amplitudes = []
        for displacement in displacements:
            cart_displacement = self._ref_structure.get_cart_displacement(displacement)
            amplitudes.append(get_amplitude(cart_basis_vector, cart_displacement))
        return (
            np.array(displacements),
            np.array(amplitudes),
            np.array(polarizabilities),
        )

    def get_masked_model(self, dof_indexes_to_mask: list[int]) -> InterpolationModel:
        """Return new model with certain degrees of freedom deactivated.

        Model masking allows for the calculation of partial Raman spectra in which only
        certain degrees of freedom are considered.
        """
        result = copy.deepcopy(self)
        new_mask = result.mask
        new_mask[:] = False
        new_mask[dof_indexes_to_mask] = True
        result.mask = new_mask
        return result

    def unmask(self) -> None:
        """Clear mask, activating all specified DOFs."""
        self._mask[:] = False

    def __repr__(self) -> str:
        """Return string representation."""
        total_dofs = 3 * len(self._ref_structure.positions)
        specified_dofs = len(self._cart_basis_vectors)
        core = f"{specified_dofs}/{total_dofs}"
        if specified_dofs == total_dofs:
            core = ANSICOLORS.OK_GREEN + core + ANSICOLORS.END
        elif 1 <= specified_dofs < total_dofs:
            core = ANSICOLORS.WARNING_YELLOW + core + ANSICOLORS.END
        else:
            core = ANSICOLORS.ERROR_RED + core + ANSICOLORS.END

        result = f"InterpolationModel with {core} degrees of freedom specified."

        num_masked = np.sum(self._mask)
        if num_masked > 0:
            num_arts = len(self._cart_basis_vectors)
            msg = f"ATTENTION: {num_masked}/{num_arts} degrees of freedom are masked."
            result += f"\n {ANSICOLORS.WARNING_YELLOW} {msg} {ANSICOLORS.END}"
        if self._is_dummy_model:
            msg = "ATTENTION: this is a dummy model."
            result += f"\n {ANSICOLORS.WARNING_YELLOW} {msg} {ANSICOLORS.END}"

        return result



"""Functions for interacting with VASP POSCAR files."""

from pathlib import Path
from typing import TextIO

import numpy as np
from numpy.typing import NDArray

from ramannoodle.io._utils import verify_structure, pathify
from ramannoodle.exceptions import InvalidFileException
from ramannoodle.constants import ATOM_SYMBOLS, ATOMIC_NUMBERS
from ramannoodle.structure._reference import ReferenceStructure
from ramannoodle.io.vasp.outcar import _get_lattice_vector_from_outcar_line

# pylint: disable=R0801


def _read_lattice(file: TextIO) -> NDArray[np.float64]:
    """Read all three lattice vectors (in angstroms) from a VASP POSCAR file.

    Raises
    ------
    InvalidFileException
    """
    try:
        file.readline()
        line = file.readline()
        scale_factor = float(line)
    except ValueError as exc:
        raise InvalidFileException(f"scale factor could not be parsed: {line}") from exc

    lattice = []
    for _ in range(3):
        try:
            line = file.readline()
            vector = _get_lattice_vector_from_outcar_line(line)
        except (EOFError, ValueError) as exc:
            raise InvalidFileException(f"lattice could not be parsed: {line}") from exc
        if vector.shape != (3,):
            raise InvalidFileException(f"lattice could not be parsed: {line}")
        lattice.append(vector)

    return np.array(lattice) * scale_factor


def _read_atomic_symbols(file: TextIO) -> list[str]:
    """Read atomic symbols from a VASP POSCAR file.

    Raises
    ------
    InvalidFileException

    """
    # First read symbols
    line = file.readline()
    symbols = line.split()
    if len(symbols) == 0:
        raise InvalidFileException("no atom symbols found")
    for symbol in symbols:
        if symbol not in ATOMIC_NUMBERS:
            raise InvalidFileException(f"unrecognized atom symbol: {symbol}")

    # Then read ion counts
    line = file.readline()
    str_counts = line.split()
    if len(str_counts) != len(symbols):
        spec = f"{len(str_counts)} != {len(symbols)}"
        raise InvalidFileException(f"wrong number of ion counts: {spec}")
    int_counts = []
    for count in str_counts:
        try:
            int_counts.append(int(count))
        except ValueError as err:
            raise InvalidFileException(f"could not parse counts: {line}") from err

    result = []
    for symbol, int_count in zip(symbols, int_counts):
        result += [symbol] * int_count
    return result


def _read_positions(
    file: TextIO, lattice: NDArray[np.float64], num_atoms: int
) -> NDArray[np.float64]:
    """Read atomic symbols from a VASP POSCAR file.

    Raises
    ------
    InvalidFileException

    """
    cart_mode = False
    label = file.readline()
    if len(label.strip()) == 0 or label[0].strip() == "":
        raise InvalidFileException(
            f"missing first character in coordinate format: '{label}'"
        )
    if label[0].lower() == "s":  # selective dynamics
        label = file.readline()
    if label[0].lower() == "c":
        cart_mode = True
    elif label[0].lower() != "d":
        raise InvalidFileException(f"unrecognized coordinate format: {label}")

    positions = []
    for _ in range(num_atoms):
        try:
            line = file.readline()
            position = [float(item) for item in line.split()[0:3]]
        except (EOFError, ValueError, IndexError) as exc:
            raise InvalidFileException(
                f"positions could not be parsed: {line}"
            ) from exc
        if len(position) != 3:
            raise InvalidFileException(f"positions could not be parsed: {line}")
        positions.append(position)

    if cart_mode:
        return np.array(positions) @ np.linalg.inv(lattice)
    return np.array(positions)


def read_positions(
    filepath: str | Path,
) -> NDArray[np.float64]:
    """Read fractional positions from a VASP POSCAR file.

    Parameters
    ----------
    filepath

    Returns
    -------
    :
        (fractional) Array with shape (N,3) where N is the number of atoms.

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        lattice = _read_lattice(file)
        atomic_symbols = _read_atomic_symbols(file)
        positions = _read_positions(file, lattice, len(atomic_symbols))
        return positions


def read_structure(
    filepath: str | Path,
) -> tuple[NDArray[np.float64], list[int], NDArray[np.float64]]:
    """Read lattice, atomic numbers, and fractional positions from a VASP POSCAR file.

    Parameters
    ----------
    filepath

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        lattice = _read_lattice(file)
        atomic_symbols = _read_atomic_symbols(file)
        atomic_numbers = [ATOMIC_NUMBERS[symbol] for symbol in atomic_symbols]
        positions = _read_positions(file, lattice, len(atomic_symbols))
        return lattice, atomic_numbers, positions


def read_ref_structure(
    filepath: str | Path,
) -> ReferenceStructure:
    """Read reference structure from a VASP POSCAR file.

    Parameters
    ----------
    filepath

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    SymmetryException
        Structural symmetry determination failed.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        lattice = _read_lattice(file)
        atomic_symbols = _read_atomic_symbols(file)
        atomic_numbers = [ATOMIC_NUMBERS[symbol] for symbol in atomic_symbols]
        positions = _read_positions(file, lattice, len(atomic_symbols))
        return ReferenceStructure(atomic_numbers, lattice, positions)


def _get_symbols_str(atomic_numbers: list[int]) -> str:
    """Return 2-line string giving elements and counts."""
    unique_numbers = []
    current_number = atomic_numbers[0]
    for number in atomic_numbers:
        if number != current_number and number in unique_numbers:
            raise ValueError(f"atomic number not grouped: {number}")
        if number != current_number:
            unique_numbers.append(current_number)
            current_number = number
    unique_numbers.append(current_number)

    result = "   " + "  ".join([ATOM_SYMBOLS[n] for n in unique_numbers]) + "\n"
    counts = [int(np.sum(np.array(atomic_numbers) == n)) for n in unique_numbers]
    result += "   " + "   ".join(str(c) for c in counts)
    return result + "\n"


def _get_lattice_str(lattice: NDArray[np.float64]) -> str:
    """Return 3-line lattice string."""
    result = ""
    for vector in lattice:
        result += f"     {vector[0]:9.16f}   {vector[1]:9.16f}   {vector[2]:9.16f}\n"
    return result


def _get_positions_str(positions: NDArray[np.float64]) -> str:
    """Return positions string."""
    result = ""
    for vector in positions:
        result += f"  {vector[0]:9.16f}   {vector[1]:9.16f}   {vector[2]:9.16f}\n"
    return result


# pylint: disable=too-many-arguments,too-many-positional-arguments
def write_structure(
    lattice: NDArray[np.float64],
    atomic_numbers: list[int],
    positions: NDArray[np.float64],
    filepath: str | Path,
    overwrite: bool = False,
    label: str = "POSCAR written by ramannoodle",
) -> None:
    """Write structure to a VASP POSCAR file.

    Parameters
    ----------
    lattice
        (Å) Array with shape (3,3).
    atomic_numbers
        List of length N where N is the number of atoms.
    positions
        (fractional) Array with shape (N,3).
    filepath
    overwrite
        Overwrite the file if it exists.
    label
        POSCAR label (first line).
    """
    verify_structure(lattice, atomic_numbers, positions)
    filepath = pathify(filepath)

    open_mode = "w" if overwrite else "x"
    filepath = pathify(filepath)

    label_str = repr(label)[1:-1] + "\n"  # Raw string with quotes removed
    lattice_str = _get_lattice_str(lattice)
    symbols_str = _get_symbols_str(atomic_numbers)
    positions_str = _get_positions_str(positions)
    with open(filepath, open_mode, encoding="utf-8") as file:
        file.write(label_str)
        file.write("   1.00000000000000" + "\n")
        file.write(lattice_str)
        file.write(symbols_str)
        file.write("Direct\n")
        file.write(positions_str)



"""Functions for interacting with files used and produced by DFT codes."""

# flake8: noqa: F401
from ramannoodle.io import generic
from ramannoodle.io import vasp
from ramannoodle.io import _utils

__all__ = ["generic", "vasp", "_utils"]



"""Exceptions, warnings, and related functions."""

from typing import Any, Sequence

from numpy.typing import NDArray


class NoMatchingLineFoundException(Exception):
    """No matching line can be found in file."""


class InvalidFileException(Exception):
    """File cannot not be read, likely due to due to invalid or unexpected format."""


class IncompatibleStructureException(Exception):
    """File contains structure that is incompatible with the current operation."""


class InvalidDOFException(Exception):
    """A supplied degree of freedom is invalid."""


class DOFWarning(UserWarning):
    """A degree of freedom may not have been specified as intended."""


class SymmetryException(Exception):
    """Symmetry operation failed."""


class UserError(Exception):
    """The user has done something they shouldn't.

    This exception is used sparingly, as (ideally) the structure of the API should
    dictate what the user should and shouldn't do.

    """


def _shape_string(shape: Sequence[int | None]) -> str:
    """Get a string representing a shape.

    Maps None --> "_", indicating that this element can be anything.
    """
    result = "("
    for i in shape:
        if i is None:
            result += "_,"
        else:
            result += f"{i},"
    if len(shape) == 1:
        return result + ")"
    return result[:-1] + ")"


def get_type_error(name: str, value: Any, correct_type: str) -> TypeError:
    """Return TypeError for an ndarray argument.

    :meta private:
    """
    wrong_type = type(value).__name__
    return TypeError(f"{name} should have type {correct_type}, not {wrong_type}")


def get_shape_error(name: str, array: NDArray[Any], desired_shape: str) -> ValueError:
    """Return ValueError for an ndarray with the wrong shape.

    :meta private:
    """
    shape_spec = f"{_shape_string(array.shape)} != {desired_shape}"
    return ValueError(f"{name} has wrong shape: {shape_spec}")


def verify_ndarray(name: str, array: NDArray[Any]) -> None:
    """Verify type of NDArray .

    :meta private: We should avoid calling this function wherever possible (EATF)
    """
    try:
        _ = array.shape
    except AttributeError as exc:
        raise get_type_error(name, array, "ndarray") from exc


def verify_ndarray_shape(
    name: str, array: NDArray[Any], shape: Sequence[int | None]
) -> None:
    """Verify an NDArray's shape.

    :meta private: We should avoid calling this function whenever possible (EATF).

    Parameters
    ----------
    shape
        int elements will be checked, None elements will not be.
    """
    try:
        if len(shape) != array.ndim:
            raise get_shape_error(name, array, _shape_string(shape))
        for d1, d2 in zip(array.shape, shape, strict=True):
            if d2 is not None and d1 != d2:
                raise get_shape_error(name, array, _shape_string(shape))
    except AttributeError as exc:
        raise get_type_error(name, array, "ndarray") from exc


def verify_list_len(name: str, array: list[Any], length: int | None) -> None:
    """Verify an list's shape.

    Calling function should be avoided whenever possible (EATF).

    :meta private:

    Parameters
    ----------
    length
        length. If None, not checked.
    """
    if not isinstance(array, list):
        raise get_type_error(name, array, "list")
    if length is not None and len(array) is not length:
        raise ValueError(f"{name} has wrong length: {len(array)} != length")


def verify_positions(name: str, array: NDArray[Any]) -> None:
    """Verify fractional positions according to dimensions and boundary conditions.

    :meta private:

    """
    verify_ndarray_shape(name, array, (None, 3))
    if (0 > array).any() or (array > 1.0).any():
        raise ValueError(f"{name} has coordinates that are not between 0 and 1")


def get_torch_missing_error() -> UserError:
    """Get error indicating that torch is not installed."""
    required_modules = "'torch', 'torch-scatter', and 'torch-sparse' packages"
    return UserError(f"torch functionality requires {required_modules}")


def get_pymatgen_missing_error() -> UserError:
    """Get error indicating that pymatgen is not installed."""
    return UserError("pymatgen functionality requires pymatgen package")



"""Polarizability dataset for pytorch."""

# flake8: noqa: F401
from ramannoodle.dataset.torch._dataset import PolarizabilityDataset
from ramannoodle.dataset.torch import utils

__all__ = ["PolarizabilityDataset", "utils"]



"""Polarizability model based on a graph neural network (GNN)."""

from __future__ import annotations

import typing

import numpy as np
from numpy.typing import NDArray

from tqdm import tqdm

from ramannoodle.structure._reference import ReferenceStructure
from ramannoodle.exceptions import (
    get_torch_missing_error,
    UserError,
    verify_ndarray_shape,
)
from ramannoodle.abstract import PolarizabilityModel

try:
    import torch
    from torch import Tensor
    from torch.nn import (
        BatchNorm1d,
        Embedding,
        Linear,
        ModuleList,
        Sequential,
        Module,
        LayerNorm,
    )
    from torch_geometric.nn.inits import reset
    from torch_geometric.nn.models.schnet import ShiftedSoftplus
    from torch_geometric.utils import scatter
    import ramannoodle.pmodel.torch._utils as rn_torch_utils
    from ramannoodle.dataset.torch.utils import (
        polarizability_vectors_to_tensors,
        polarizability_tensors_to_vectors,
    )

except (ModuleNotFoundError, UserError) as exc:
    raise get_torch_missing_error() from exc

# pylint: disable=not-callable


class _GaussianFilter(torch.nn.Module):
    """Gaussian filter.

    Parameters should be chosen such that all expected inputs are between lower and
    upper bounds.

    Parameters
    ----------
    lower_bound
    upper_bound
    steps
        Number of steps to take between ``lower_bound`` and ``upper_bound``.
    """

    def __init__(self, lower_bound: float, upper_bound: float, steps: int):
        super().__init__()
        offset = torch.linspace(lower_bound, upper_bound, steps)
        self.coefficient = -0.5 / (float(offset[1]) - float(offset[0])) ** 2
        self.register_buffer("offset", offset)

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        x
            Tensor with size [D,]. Typically contains interatomic distances.

        Returns
        -------
        :
            Tensor with size [D,steps].

        """
        x = x.view(-1, 1) - self.offset.view(1, -1)
        return torch.exp(self.coefficient * x.pow(2))


class _NodeBlock(torch.nn.Module):
    """Edge to node message passer.

    Architecture and notation is based on equation (5) in https://doi.org/10.1038/
    s41524-021-00543-3. The architecture has been modified to use layer normalization.

    Parameters
    ----------
    size_node_embedding
    size_edge_embedding

    """

    def __init__(
        self,
        size_node_embedding: int,
        size_edge_embedding: int,
    ):
        super().__init__()

        # Combination of two linear layers, dubbed "core" and "filter":
        #   "filter" : c1ij W1 + b1
        #   "core"   : c1ij W2 + b2
        self.c1_linear = Linear(
            size_node_embedding + size_edge_embedding,
            2 * size_node_embedding,
        )

        self.c1_norm = LayerNorm(2 * size_node_embedding)
        self.final_norm = LayerNorm(size_node_embedding)

    def reset_parameters(self) -> None:
        """Reset model parameters."""
        self.c1_linear.reset_parameters()
        self.c1_norm.reset_parameters()
        self.final_norm.reset_parameters()

    def forward(
        self, node_embedding: Tensor, edge_embedding: Tensor, i: Tensor
    ) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        node_embedding
            Tensor with size [N,size_node_embedding] where N is the number of
            nodes.
        edge_embedding
            Tensor with size [E,size_edge_embedding] where E is the number of
            edges.

        Returns
        -------
        :
            Tensor with size [N,size_node_embedding].
        """
        c1 = torch.cat([node_embedding[i], edge_embedding], dim=1)
        c1 = self.c1_norm(self.c1_linear(c1))
        c1_filter, c1_core = c1.chunk(2, dim=1)
        c1_filter = c1_filter.sigmoid()
        c1_core = c1_core.tanh()
        c1_emb = scatter(
            c1_filter * c1_core, i, dim=0, dim_size=node_embedding.size(0), reduce="sum"
        )
        c1_emb = self.final_norm(c1_emb)

        return typing.cast(Tensor, (node_embedding + c1_emb).tanh())


class _EdgeBlock(torch.nn.Module):
    """Node to edge message passer.

    Architecture and notation is based on equation (6) in https://doi.org/10.1038/
    s41524-021-00543-3. The architecture has been modified to use batch normalization.

    Parameters
    ----------
    size_node_embedding
    size_edge_embedding
    """

    def __init__(
        self,
        size_node_embedding: int,
        size_edge_embedding: int,
    ):
        super().__init__()

        # Combination of two linear layers, dubbed "core" and "filter":
        #   "filter" : c2ij W3 + b3
        #   "core"   : c2ij W4 + b4
        self.c2_linear = Linear(size_node_embedding, 2 * size_edge_embedding)

        # Combination of two linear layers, dubbed "core" and "filter":
        #   "filter" : c3ij W5 + b5
        #   "core"   : c3ij W6 + b6
        self.c3_linear = Linear(
            3 * size_node_embedding + 2 * size_edge_embedding,
            2 * size_edge_embedding,
        )

        self.c2_norm_1 = LayerNorm(2 * size_edge_embedding)
        self.c3_norm_1 = LayerNorm(2 * size_edge_embedding)
        self.c2_norm_2 = LayerNorm(size_edge_embedding)
        self.c3_norm_2 = LayerNorm(size_edge_embedding)

    def reset_parameters(self) -> None:
        """Reset model parameters."""
        self.c2_linear.reset_parameters()
        self.c3_linear.reset_parameters()
        self.c2_norm_1.reset_parameters()
        self.c3_norm_1.reset_parameters()
        self.c2_norm_2.reset_parameters()
        self.c3_norm_2.reset_parameters()

    def _get_c2_embedding(
        self,
        node_embedding: Tensor,
        i: Tensor,
        j: Tensor,
    ) -> Tensor:
        """Get c2 embedding.

        Parameters
        ----------
        node_embedding
            Tensor with size [N,size_node_embedding] where N is the number of
            nodes.
        i
            Node 1 of edge pairs, a tensor with size [E,].
        j
            Node 2 of edge pairs, a tensor with size [E,].

        Returns
        -------
        :
            Tensor with size [E,size_edge_embedding].
        """
        c2 = node_embedding[i] * node_embedding[j]
        c2 = self.c2_norm_1(self.c2_linear(c2))
        c2_filter, c2_core = c2.chunk(2, dim=1)
        c2_filter = c2_filter.sigmoid()
        c2_core = c2_core.tanh()
        return typing.cast(Tensor, self.c2_norm_2(c2_filter * c2_core))

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def _get_c3_embedding(
        self,
        node_embedding: Tensor,
        edge_embedding: Tensor,
        index_i: Tensor,
        index_j: Tensor,
        index_k: Tensor,
        index_ji: Tensor,
        index_kj: Tensor,
    ) -> Tensor:
        """Get c3 embedding.

        Parameters
        ----------
        node_embedding
            Tensor with size [N,size_node_embedding] where N is the number of
            nodes.
        edge_embedding
            Tensor with size [E,size_edge_embedding] where E is the number of
            edges.
        index_i
            Node 1 of edge triplets, a tensor with size [T,] where T is the number
            of triplets.
        index_j
            Node 2 of edge triplets, a tensor with size [T,].
        index_k
            Node 3 of edge triplets, a tensor with size [T,].
        index_ji
            Index of (j,i) corresponding to (index_j,index_i), a tensor with size
            [T,].
        index_kj
            Index of (k,j) corresponding to (index_k,index_j), a tensor with size
            [T,].

        Returns
        -------
        :
            Tensor with size [E,size_edge_embedding].
        """
        c3 = torch.cat(
            [
                node_embedding[index_i],
                node_embedding[index_j],
                node_embedding[index_k],
                edge_embedding[index_ji],
                edge_embedding[index_kj],
            ],
            dim=1,
        )
        c3 = self.c3_norm_1(self.c3_linear(c3))
        c3_filter, c3_core = c3.chunk(2, dim=1)
        c3_filter = c3_filter.sigmoid()
        c3_core = c3_core.tanh()
        c3_emb = scatter(
            c3_filter * c3_core,
            index_ji,
            dim=0,
            dim_size=edge_embedding.size(0),
            reduce="sum",
        )
        return typing.cast(Tensor, self.c3_norm_2(c3_emb))

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def forward(
        self,
        node_embedding: Tensor,
        edge_embedding: Tensor,
        i: Tensor,
        j: Tensor,
        index_i: Tensor,
        index_j: Tensor,
        index_k: Tensor,
        index_ji: Tensor,
        index_kj: Tensor,
    ) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        node_embedding
            Tensor with size [N,size_node_embedding] where N is the number of
            nodes.
        edge_embedding
            Tensor with size [E,size_edge_embedding] where E is the number of
            edges.
        i
            Node 1 of edge pairs, a tensor with size [E,].
        j
            Node 2 of edge pairs, a tensor with size [E,].
        index_i
            Node 1 of edge triplets, a tensor with size [T,] where T is the number
            of triplets.
        index_j
            Node 2 of edge triplets, a tensor with size [T,].
        index_k
            Node 3 of edge triplets, a tensor with size [T,].
        index_ji
            Index of (j,i) corresponding to (index_j,index_i), a tensor with size
            [T,].
        index_kj
            Index of (k,j) corresponding to (index_k,index_j), a tensor with size
            [T,].

        Returns
        -------
        :
            Tensor with size [E,size_edge_embedding].

        """
        c2_embedding = self._get_c2_embedding(node_embedding, i, j)
        # pylint: disable=duplicate-code
        c3_embedding = self._get_c3_embedding(
            node_embedding,
            edge_embedding,
            index_i,
            index_j,
            index_k,
            index_ji,
            index_kj,
        )
        return (edge_embedding + c2_embedding + c3_embedding).tanh()


def _get_edge_polarizability_vectors(
    polarizability_embedding: Tensor,
    unit_vector: Tensor,
) -> Tensor:
    """Convert polarizability embedding to edge polarizability vectors.

    Parameters
    ----------
    polarizability_embedding
        Tensor with size [E,12] where E is the number of edges.
    unit_vector
        (Å) Unit vectors of edges, a tensor with size [E,3].

    Returns
    -------
    :
        Tensor with size [E,6].
    """
    a1 = torch.zeros((polarizability_embedding.size(0), 3, 3))
    a1[:, 0, 0] = polarizability_embedding[:, 0]
    a1[:, 1, 1] = polarizability_embedding[:, 1]
    a1[:, 2, 2] = polarizability_embedding[:, 1]
    a2 = torch.zeros((polarizability_embedding.size(0), 3, 3))
    a2[:, 0, 0] = polarizability_embedding[:, 2]
    a2[:, 1, 1] = polarizability_embedding[:, 3]
    a2[:, 2, 2] = polarizability_embedding[:, 3]
    a3 = torch.zeros((polarizability_embedding.size(0), 3, 3))
    a3[:, 0, 0] = polarizability_embedding[:, 4]
    a3[:, 1, 1] = polarizability_embedding[:, 5]
    a3[:, 2, 2] = polarizability_embedding[:, 5]
    a4 = torch.zeros((polarizability_embedding.size(0), 3, 3))
    a4[:, 0, 0] = polarizability_embedding[:, 6]
    a4[:, 1, 1] = polarizability_embedding[:, 7]
    a4[:, 2, 2] = polarizability_embedding[:, 7]
    a5 = torch.zeros((polarizability_embedding.size(0), 3, 3))
    a5[:, 0, 0] = polarizability_embedding[:, 8]
    a5[:, 1, 1] = polarizability_embedding[:, 9]
    a5[:, 2, 2] = polarizability_embedding[:, 9]
    a6 = torch.zeros((polarizability_embedding.size(0), 3, 3))
    a6[:, 0, 0] = polarizability_embedding[:, 10]
    a6[:, 1, 1] = polarizability_embedding[:, 11]
    a6[:, 2, 2] = polarizability_embedding[:, 11]

    rotation = rn_torch_utils.get_rotations(unit_vector)
    inv_rotation = torch.linalg.inv(rotation)

    a1 = rotation @ a1 @ inv_rotation
    a2 = rotation @ a2 @ inv_rotation
    a3 = rotation @ a3 @ inv_rotation
    a4 = rotation @ a4 @ inv_rotation
    a5 = rotation @ a5 @ inv_rotation
    a6 = rotation @ a6 @ inv_rotation

    edge_polarizability = (
        a1 * torch.tensor([[0, 1, 0], [1, 0, 0], [0, 0, 0]])
        + a2 * torch.tensor([[0, 0, 1], [0, 0, 0], [1, 0, 0]])
        + a3 * torch.tensor([[0, 0, 0], [0, 0, 1], [0, 1, 0]])
        + a4 * torch.tensor([[1, 0, 0], [0, 0, 0], [0, 0, 0]])
        + a5 * torch.tensor([[0, 0, 0], [0, 1, 0], [0, 0, 0]])
        + a6 * torch.tensor([[0, 0, 0], [0, 0, 0], [0, 0, 1]])
    )
    return polarizability_tensors_to_vectors(edge_polarizability)


class PotGNN(
    Module, PolarizabilityModel
):  # pylint: disable=too-many-instance-attributes
    r"""POlarizability Tensor Graph Neural Network (PotGNN).

    The architecture was inspired by the "direct force architecture" developed in Park
    `et al.`;  `npj Computational Materials` (2021)7:73;
    `doi:10.1038/s41524-021-00543-3 <https://doi.org/10.1038/s41524-021-00543-3>`_.
    Implementation adapted from ``torch_geometric.nn.models.GNNFF``
    authored by @ken2403 and merged by @rusty1s.

    The architecture of this model is still somewhat in flux. More complete
    documentation for this model, including a description of the architecture and
    discussion of design choices, will be available at a later date.

    Parameters
    ----------
    ref_structure
        Reference structure from which nodes/edges are determined.
    cutoff
        (Å) Cutoff distance for edges.
    size_node_embedding
    size_edge_embedding
    num_message_passes
    gaussian_filter_start
        (Å) Lower bound of the Gaussian filter used in initial edge embedding.
    gaussian_filter_end
        (Å) Upper bound of the Gaussian filter used in initial edge embedding.
    mean_polarizability
        Array with shape (3,3).
    stddev_polarizability
        Array with shape (3,3).
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals
    def __init__(
        self,
        ref_structure: ReferenceStructure,
        cutoff: float,
        size_node_embedding: int,
        size_edge_embedding: int,
        num_message_passes: int,
        gaussian_filter_start: float,
        gaussian_filter_end: float,
        mean_polarizability: NDArray[np.float64],
        stddev_polarizability: NDArray[np.float64],
    ):
        super().__init__()
        # Validate parameters
        if cutoff <= 0:
            raise ValueError(f"invalid cutoff: {cutoff} <= 0")
        if size_node_embedding <= 0:
            raise ValueError(f"invalid size_node_embedding: {size_node_embedding} <= 0")
        if size_edge_embedding <= 0:
            raise ValueError(f"invalid size_edge_embedding: {size_edge_embedding} <= 0")
        if num_message_passes <= 0:
            raise ValueError(f"invalid num_message_passes: {num_message_passes} <= 0")
        if gaussian_filter_start < 0:
            inequality = f"{gaussian_filter_start} < 0"
            raise ValueError(f"invalid gaussian_filter_start: {inequality}")
        if gaussian_filter_end <= gaussian_filter_start:
            inequality = f"{gaussian_filter_end} <= gaussian_filter_start"
            raise ValueError(f"invalid gaussian_filter_end: {inequality}")
        verify_ndarray_shape("mean_polarizability", mean_polarizability, (3, 3))
        verify_ndarray_shape("stddev_polarizability", stddev_polarizability, (3, 3))

        self._ref_structure = ref_structure
        self._cutoff = cutoff
        self._mean_polarizability = mean_polarizability
        self._stddev_polarizability = stddev_polarizability

        # Set up graph.
        lattice = torch.from_numpy(ref_structure.lattice).unsqueeze(0)
        positions = torch.from_numpy(ref_structure.positions).unsqueeze(0)
        self._ref_edge_indexes, _, _ = rn_torch_utils._radius_graph_pbc(
            lattice.type(torch.get_default_dtype()).to(torch.get_default_device()),
            positions.type(torch.get_default_dtype()).to(torch.get_default_device()),
            cutoff,
        )
        self._batch_triplets = rn_torch_utils.BatchTriplets(
            len(ref_structure.atomic_numbers), self._ref_edge_indexes
        )

        # Atom types map
        unique_atomic_numbers = set(ref_structure.atomic_numbers)
        self._num_atom_types = len(unique_atomic_numbers)
        self._atom_type_map = (torch.zeros(119) - 1).type(torch.int)
        for atom_type, atomic_number in enumerate(unique_atomic_numbers):
            self._atom_type_map[atomic_number] = atom_type

        self._node_embedding = Sequential(
            Embedding(self._num_atom_types, size_node_embedding),
            ShiftedSoftplus(),
            Linear(size_node_embedding, size_node_embedding),
            ShiftedSoftplus(),
            Linear(size_node_embedding, size_node_embedding),
        )
        self._edge_embedding = _GaussianFilter(
            gaussian_filter_start, gaussian_filter_end, size_edge_embedding
        )

        self._node_blocks = ModuleList(
            [
                _NodeBlock(size_node_embedding, size_edge_embedding)
                for _ in range(num_message_passes)
            ]
        )
        self._edge_blocks = ModuleList(
            [
                _EdgeBlock(size_node_embedding, size_edge_embedding)
                for _ in range(num_message_passes)
            ]
        )

        self._to_polarizability_embedding = Sequential(
            Linear(size_edge_embedding, size_edge_embedding),
            BatchNorm1d(size_edge_embedding),
            ShiftedSoftplus(),
            Linear(size_edge_embedding, size_edge_embedding),
            ShiftedSoftplus(),
            Linear(size_edge_embedding, 12),
        )

    def _convert_to_atom_type(self, atomic_numbers: Tensor) -> Tensor:
        """Convert atomic numbers into atom types.

        Atom types must be defined for all atomic numbers.

        Parameters
        ----------
        atomic_numbers
            Tensor with arbitrary shape.

        Returns
        -------
        :
            Tensor with the same shape as atomic_numbers.

        """
        return self._atom_type_map[atomic_numbers]

    def reset_parameters(self) -> None:
        """Reset model parameters."""
        reset(self._node_embedding)
        for node_block in self._node_blocks:
            node_block.reset_parameters()
        for edge_block in self._edge_blocks:
            edge_block.reset_parameters()
        reset(self._to_polarizability_embedding)

    def _batch_graph(
        self, lattice: Tensor, positions: Tensor
    ) -> tuple[Tensor, Tensor, Tensor]:
        """Generate edge indexes, unit vectors, and edge distances for a batch.

        Parameters
        ----------
        lattice
            (Å) Tensor with size [S,3,3] where S is the number of samples.
        positions
            (fractional) Tensor with size [S,N,3] where N is the number of atoms.

        Returns
        -------
        :
            0.  edge indexes -- Tensor of size [3,E] where E is the number of edges. The
                first element are the graph indexes, while the remaining two elements
                are edge indexes.

            #.  unit vectors -- (Å) Tensor with size [E,3].

            #.  distances -- (Å) Tensor with size [E,].

        """
        num_samples = lattice.size(0)
        num_atoms = positions.size(1)

        # Get edge indexes
        graph_indexes = torch.tensor([range(num_samples)]).repeat_interleave(
            self._ref_edge_indexes.size(1), dim=1
        )
        edge_indexes = self._ref_edge_indexes.repeat((1, num_samples))
        edge_indexes[0] = graph_indexes

        # Compute pairwise distance matrix.
        displacement = positions.unsqueeze(1) - positions.unsqueeze(2)
        displacement = torch.where(
            displacement % 1 > 0.5, displacement % 1 - 1, displacement % 1
        )
        expanded_lattice = lattice[:, None, :, :].expand(
            -1, displacement.size(1), -1, -1
        )
        cart_displacement = displacement.matmul(expanded_lattice)
        cart_distance_matrix = torch.sqrt(torch.sum(cart_displacement**2, dim=-1))

        return rn_torch_utils.get_graph_info(
            cart_displacement, edge_indexes, cart_distance_matrix, num_atoms
        )

    def forward(  # pylint: disable=too-many-locals
        self,
        lattice: Tensor,
        atomic_numbers: Tensor,
        positions: Tensor,
    ) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        lattice
            (Å) Tensor with size [S,3,3] where S is the number of samples.
        atomic_numbers
            Tensor with size [S,N] where N is the number of atoms.
        positions
            (fractional) Tensor with size [S,N,3].

        Returns
        -------
        :
            Polarizability vectors with size [S,6]. To convert into tensor form, see
            :func:`polarizability_vectors_to_tensors`.

        """
        edge_index, unit_vector, distance = self._batch_graph(lattice, positions)
        atom_types = self._convert_to_atom_type(atomic_numbers).flatten()
        node_emb = self._node_embedding(atom_types)
        edge_emb = self._edge_embedding(distance)

        # Message passing blocks:
        triplets = self._batch_triplets.get_triplets(batch_size=lattice.size(0))
        for node_block, edge_block in zip(self._node_blocks, self._edge_blocks):
            node_emb = node_block(node_emb, edge_emb, triplets[0])
            edge_emb = edge_block(node_emb, edge_emb, *triplets)

        # Get edge polarizabilities.
        polarizability_emb = self._to_polarizability_embedding(edge_emb)
        edge_polarizability = _get_edge_polarizability_vectors(
            polarizability_emb, unit_vector
        )

        # Isolate polarizabilities from batch graphs.
        polarizability = torch.zeros((positions.size(0), 6))
        for structure_i in range(positions.size(0)):
            mask = (edge_index[0:1] == structure_i).T
            count = torch.sum(mask)
            polarizability[structure_i] = torch.sum(edge_polarizability * mask, dim=0)
            polarizability[structure_i] /= count
        return polarizability

    def calc_polarizabilities(
        self, positions_batch: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Return estimated polarizabilities for a batch of fractional positions.

        Parameters
        ----------
        positions_batch
            (fractional) Array with shape (S,N,3) where S is the number of samples
            and N is the number of atoms.

        Returns
        -------
        :
            Array with shape (S,3,3).
        """
        verify_ndarray_shape(
            "positions_batch", positions_batch, (None, self._ref_structure.num_atoms, 3)
        )
        self.eval()

        polarizabilities = torch.zeros((positions_batch.shape[0], 3, 3))
        positions_subbatchs = rn_torch_utils.batch_positions(
            positions_batch, batch_size=100
        )
        progress_bar = tqdm(total=positions_batch.shape[0], unit=" configs")

        start_index = 0
        for positions_subbatch in positions_subbatchs:
            subbatch_size = positions_subbatch.shape[0]

            lattice = torch.tensor(self._ref_structure.lattice)
            lattice = lattice.unsqueeze(0).expand((subbatch_size, -1, -1))
            atomic_numbers = torch.tensor(
                self._ref_structure.atomic_numbers, dtype=torch.int
            )
            atomic_numbers = atomic_numbers.unsqueeze(0).expand((subbatch_size, -1))

            default_type = torch.get_default_dtype()
            polarizability = self.forward(
                lattice.type(default_type),
                atomic_numbers,
                torch.tensor(positions_subbatch).type(default_type),
            )
            polarizability = polarizability_vectors_to_tensors(polarizability)
            end_index = start_index + subbatch_size
            polarizabilities[start_index:end_index] = polarizability.detach()
            progress_bar.update(end_index - start_index)
            start_index = end_index

        progress_bar.close()

        result = polarizabilities.detach().numpy() * self._stddev_polarizability
        result += self._mean_polarizability
        return result



"""Polarizability PyTorch dataset."""

import numpy as np
from numpy.typing import NDArray

from ramannoodle.exceptions import (
    verify_ndarray_shape,
    verify_list_len,
    get_torch_missing_error,
)

try:
    import torch
    from torch import Tensor
    from torch.utils.data import Dataset
    import ramannoodle.dataset.torch.utils as rn_torch_utils
except ModuleNotFoundError as exc:
    raise get_torch_missing_error() from exc


def _scale_and_flatten_polarizabilities(
    polarizabilities: Tensor,
    scale_mode: str,
) -> tuple[Tensor, Tensor, Tensor]:
    """Scale and flatten polarizabilities.

    3x3 polarizabilities are flattened into 6-vectors: (xx,yy,zz,xy,xz,yz).

    Parameters
    ----------
    polarizabilities
        Tensor with size [S,3,3] where S is the number of samples.
    scale_mode
        Supports ``"standard"`` (standard scaling), ``"stddev"`` (division by
        standard deviation), and ``"none"`` (no scaling).

    Returns
    -------
    :
        0.  mean -- Element-wise mean of polarizabilities.
        #.  standard deviation -- Element-wise standard deviation of polarizabilities.
        #.  polarizability vectors -- Tensor with size [S,6].

    """
    mean = polarizabilities.mean(0, keepdim=True)
    stddev = polarizabilities.std(0, unbiased=False, keepdim=True)
    if scale_mode == "standard":
        polarizabilities = (polarizabilities - mean) / stddev
    elif scale_mode == "stddev":
        polarizabilities = (polarizabilities - mean) / stddev + mean
    elif scale_mode != "none":
        raise ValueError(f"unsupported scale mode: {scale_mode}")

    scaled_polarizabilities = torch.zeros((polarizabilities.size(0), 6))
    scaled_polarizabilities[:, 0] = polarizabilities[:, 0, 0]
    scaled_polarizabilities[:, 1] = polarizabilities[:, 1, 1]
    scaled_polarizabilities[:, 2] = polarizabilities[:, 2, 2]
    scaled_polarizabilities[:, 3] = polarizabilities[:, 0, 1]
    scaled_polarizabilities[:, 4] = polarizabilities[:, 0, 2]
    scaled_polarizabilities[:, 5] = polarizabilities[:, 1, 2]

    return mean, stddev, scaled_polarizabilities


class PolarizabilityDataset(Dataset[tuple[Tensor, Tensor, Tensor, Tensor]]):
    """PyTorch dataset of atomic structures and polarizabilities.

    Polarizabilities are standard scaled and flattened into 6-vectors containing the
    independent tensor components.

    Parameters
    ----------
    lattice
        (Å) Array with shape (3,3).
    atomic_numbers
        List of length N where N is the number of atoms.
    positions
        (fractional) Array with shape (S,N,3) where S is the number of samples.
    polarizabilities
        Array with shape (S,3,3).
    scale_mode
        Supports ``"standard"`` (standard scaling), ``"stddev"`` (division by
        standard deviation), and ``"none"`` (no scaling).

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        lattice: NDArray[np.float64],
        atomic_numbers: list[int],
        positions: NDArray[np.float64],
        polarizabilities: NDArray[np.float64],
    ):
        # Validate parameter shapes
        verify_ndarray_shape("lattice", lattice, (3, 3))
        verify_list_len("atomic_numbers", atomic_numbers, None)
        num_atoms = len(atomic_numbers)
        verify_ndarray_shape("positions", positions, (None, num_atoms, 3))
        num_samples = positions.shape[0]
        verify_ndarray_shape("polarizabilities", polarizabilities, (num_samples, 3, 3))

        default_type = torch.get_default_dtype()
        self._lattices = torch.tensor(lattice).type(default_type).unsqueeze(0)
        self._lattices = self._lattices.expand(num_samples, 3, 3)
        self._atomic_numbers = torch.tensor(atomic_numbers).type(torch.int).unsqueeze(0)
        self._atomic_numbers = self._atomic_numbers.expand(num_samples, num_atoms)
        self._positions = torch.tensor(positions).type(default_type)
        self._polarizabilities = torch.tensor(polarizabilities)

        _, _, scaled = _scale_and_flatten_polarizabilities(
            self._polarizabilities, scale_mode="standard"
        )
        self._scaled_polarizabilities = scaled.type(default_type)

    @property
    def num_atoms(self) -> int:
        """Get number of atoms per sample."""
        return self._positions.size(1)

    @property
    def num_samples(self) -> int:
        """Get number of samples."""
        return self._positions.size(0)

    @property
    def atomic_numbers(self) -> list[int]:
        """Get (a copy of) atomic numbers.

        Returns
        -------
        :
            List of length N where N is the number of atoms.
        """
        return [int(n) for n in self._atomic_numbers[0]]

    @property
    def positions(self) -> NDArray[np.float64]:
        """Get (a copy of) positions.

        Returns
        -------
        :
            Array with shape (S,N,3) where S is the number of samples and N is the
            number of atoms.
        """
        return self._positions.detach().clone().numpy()

    @property
    def polarizabilities(self) -> NDArray[np.float64]:
        """Get (a copy of) polarizabilities.

        Returns
        -------
        :
            Array with shape (S,3,3) where S is the number of samples.
        """
        return self._polarizabilities.detach().clone().numpy()

    @property
    def scaled_polarizabilities(self) -> NDArray[np.float64]:
        """Get (a copy of) scaled polarizabilities.

        Returns
        -------
        :
            Array with shape (S,6) where S is the number of samples.
        """
        return self._scaled_polarizabilities.detach().clone().numpy()

    @property
    def mean_polarizability(self) -> NDArray[np.float64]:
        """Get mean polarizability.

        Return
        ------
        :
            Array with shape (3,3).
        """
        return self._polarizabilities.mean(0).clone().numpy()

    @property
    def stddev_polarizability(self) -> NDArray[np.float64]:
        """Get standard deviation of polarizabilities.

        Return
        ------
        :
            Array with shape (3,3).
        """
        result = self._polarizabilities.std(0, unbiased=False)
        return result.clone().numpy()

    def scale_polarizabilities(
        self, mean: NDArray[np.float64], stddev: NDArray[np.float64]
    ) -> None:
        """Standard-scale polarizabilities given a mean and standard deviation.

        This method may be used to scale validation or test datasets according
        to the mean and standard deviation of the training set, as is best practice.

        Parameters
        ----------
        mean
            Array with shape (3,3).
        stddev
            Array with shape (3,3).

        """
        verify_ndarray_shape("mean", mean, (3, 3))
        verify_ndarray_shape("mean", stddev, (3, 3))

        scaled = self._polarizabilities.detach().clone()
        scaled = scaled - torch.tensor(mean)
        scaled /= torch.tensor(stddev)
        self._scaled_polarizabilities = (
            rn_torch_utils.polarizability_tensors_to_vectors(scaled)
        )

    def __len__(self) -> int:
        """Get number of samples."""
        return self.num_samples

    def __getitem__(self, i: int) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        """Get lattice, atomic numbers, positions, and scaled polarizabilities."""
        return (
            self._lattices[i],
            self._atomic_numbers[i],
            self._positions[i],
            self._scaled_polarizabilities[i],
        )



"""Reference structure."""

import copy

import numpy as np
from numpy.typing import NDArray
import spglib

from ramannoodle.exceptions import (
    SymmetryException,
    get_type_error,
    verify_ndarray_shape,
    verify_list_len,
    get_shape_error,
)
from ramannoodle.structure.utils import (
    displace_positions,
    transform_positions,
    apply_pbc_displacement,
    calc_displacement,
)
from ramannoodle.constants import ATOM_SYMBOLS
from ramannoodle.structure import _symmetry_utils


def _compute_permutation_matrices(
    rotations: NDArray[np.float64],
    translations: NDArray[np.float64],
    positions: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Expresses a series of rotation/translations as permutation matrices."""
    # Ensure no atom is at unit cell boundary by shifting
    # center of mass
    center_of_mass_shift = np.array([0.5, 0.5, 0.5]) - np.mean(positions, axis=0)

    permutation_matrices = []
    for rotation, translation in zip(rotations, translations):
        permutation_matrices.append(
            _get_positions_permutation_matrix(
                displace_positions(positions, center_of_mass_shift),
                transform_positions(
                    positions, rotation, translation + center_of_mass_shift
                ),
            )
        )
    return np.array(permutation_matrices)


def _get_positions_permutation_matrix(
    reference_positions: NDArray[np.float64], permuted_positions: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Calculate a permutation matrix between reference and permuted positions.

    Parameters
    ----------
    reference_positions
        (fractional) Array with shape (N,3)
    permuted_positions
        (fractional) Array with shape (N,3).

    """
    # Compute pairwise distance matrix.
    reference_positions = np.expand_dims(reference_positions, 0)
    permuted_positions = np.expand_dims(permuted_positions, 1)
    displacement = reference_positions - permuted_positions
    displacement = np.where(
        displacement % 1 > 0.5, displacement % 1 - 1, displacement % 1
    )
    distance_matrix = np.sqrt(np.sum(displacement**2, axis=-1))
    permutation_matrix = (distance_matrix < 1e-5).T

    return permutation_matrix


class ReferenceStructure:
    """Reference crystal structure, typically used by polarizability models.

    Parameters
    ----------
    atomic_numbers
        List of length N where N is the number of atoms.
    lattice
        (Å) Array with shape (3,3).
    positions
        (fractional) Array with shape (N,3).
    symprec
        (Å) Distance tolerance for symmetry search (spglib).
    angle_tolerance
        (°) Angle tolerance for symmetry search (spglib).

    Raises
    ------
    SymmetryException
        Structural symmetry determination failed.

    """

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        atomic_numbers: list[int],
        lattice: NDArray[np.float64],
        positions: NDArray[np.float64],
        symprec: float = 1e-5,
        angle_tolerance: float = -1.0,
    ) -> None:
        verify_list_len("atomic_numbers", atomic_numbers, None)
        verify_ndarray_shape("lattice", lattice, (3, 3))
        verify_ndarray_shape("positions", positions, (len(atomic_numbers), 3))

        self._atomic_numbers = atomic_numbers
        self._lattice = lattice
        self._positions = positions

        cell = (list(lattice), list(positions), atomic_numbers)
        self._symmetry_dict: dict[str, NDArray[np.float64]] | None = (
            spglib.get_symmetry(cell, symprec=symprec, angle_tolerance=angle_tolerance)
        )
        if self._symmetry_dict is None:
            raise SymmetryException("Symmetry search failed. Check structure.")

        self._rotations = self._symmetry_dict["rotations"]
        self._translations = self._symmetry_dict["translations"]
        self._permutation_matrices = _compute_permutation_matrices(
            self._rotations, self._translations, self._positions
        )

    @property
    def atomic_numbers(self) -> list[int]:
        """Get (a copy of) atomic numbers."""
        return copy.copy(self._atomic_numbers)

    @property
    def num_atoms(self) -> int:
        """Get number of atoms."""
        return len(self._atomic_numbers)

    @property
    def lattice(self) -> NDArray[np.float64]:
        """Get (a copy of) lattice.

        Returns
        -------
        :
            Å | Array with shape (3,3).
        """
        return self._lattice.copy()

    @property
    def positions(self) -> NDArray[np.float64]:
        """Get (a copy of) fractional positions.

        Returns
        -------
        :
            (fractional) Array with shape (N,3) where N is the number of atoms.
        """
        return self._positions.copy()

    @property
    def num_nonequivalent_atoms(self) -> int:
        """Get number of nonequivalent atoms."""
        assert self._symmetry_dict is not None
        return len(set(self._symmetry_dict["equivalent_atoms"]))

    def get_equivalent_atom_dict(self) -> dict[int, list[int]]:
        """Get dictionary of equivalent atoms indexes.

        Returns
        -------
        :
            atom index --> list of equivalent atom indexes

        """
        assert self._symmetry_dict is not None
        result: dict[int, list[int]] = {}
        for index, equiv_index in enumerate(self._symmetry_dict["equivalent_atoms"]):
            if index == equiv_index:
                result[index] = []
            else:
                result[equiv_index].append(index)
        return result

    def get_equivalent_displacements(
        self, displacement: NDArray[np.float64]
    ) -> list[dict[str, list[NDArray[np.float64]]]]:
        """Calculate symmetrically equivalent displacements.

        Parameters
        ----------
        displacement
            (fractional) Array with shape (N,3) where N is the number of atoms.

        Returns
        -------
        :
            List of dictionaries containing displacements and transformations,
            accessed using the 'displacements' and 'transformations' keys. Displacements
            within each dictionary will be collinear, corresponding to
            the same degree of freedom. The provided transformations are those that
            transform ``displacement`` into that degree of freedom. Displacements are
            in fractional coordinates.

        """
        displacement = apply_pbc_displacement(displacement)
        # Scale the displacement for numerical reasons.
        displacement = displacement / (np.linalg.norm(displacement) * 10)

        ref_positions = displace_positions(self._positions, displacement)

        result = []
        orthogonal_displacements: list[NDArray[np.float64]] = []
        for rotation, translation, permutation_matrix in zip(
            self._rotations, self._translations, self._permutation_matrices
        ):

            # Transform, permute, then get candidate displacement
            candidate_positions = transform_positions(
                ref_positions, rotation, translation
            )
            candidate_positions = permutation_matrix @ candidate_positions
            candidate_displacement = calc_displacement(
                self._positions,
                candidate_positions,
            )

            orthogonal_result = _symmetry_utils.is_orthogonal_to_all(
                candidate_displacement.flatten(),
                [item.flatten() for item in orthogonal_displacements],
            )

            # If new orthogonal displacement discovered
            if orthogonal_result == -1:
                result.append(
                    {
                        "displacements": [candidate_displacement],
                        "transformations": [(rotation, translation)],
                    }
                )
                orthogonal_displacements.append(candidate_displacement)
            else:
                # Candidate can be collinear to maximum of one orthogonal vector
                collinear_result = _symmetry_utils.is_non_collinear_with_all(
                    candidate_displacement,
                    orthogonal_displacements,
                )
                if collinear_result != -1:  # collinear
                    collinear_displacements = result[collinear_result]["displacements"]

                    # Check if we have a duplicate
                    is_duplicate = False
                    for collinear_displacement in collinear_displacements:
                        if np.isclose(
                            candidate_displacement, collinear_displacement, atol=1e-05
                        ).all():
                            is_duplicate = True
                            break
                    # If unique, add
                    if not is_duplicate:
                        result[collinear_result]["displacements"].append(
                            candidate_displacement
                        )
                        result[collinear_result]["transformations"].append(
                            (rotation, translation)
                        )

        return result

    def get_cart_displacement(
        self, displacement: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Convert a (fractional) displacement into Cartesian coordinates.

        Parameters
        ----------
        displacement
            (fractional) Array with shape (...,N,3)  where N is the number of atoms.

        Returns
        -------
        :
            (Å) Array with shape (...,N,3).
        """
        displacement = apply_pbc_displacement(displacement)

        return displacement @ self._lattice

    def get_cart_direction(self, direction: NDArray[np.float64]) -> NDArray[np.float64]:
        """Convert a (fractional) direction into cartesian coordinates.

        Parameters
        ----------
        direction
            (fractional) Array with shape (3,).

        Returns
        -------
        :
            (Å) Array with shape (3,).
        """
        direction = apply_pbc_displacement(direction)
        try:
            return np.array([direction]) @ self._lattice
        except ValueError as exc:
            raise get_shape_error("direction", direction, "(3,)") from exc

    def get_frac_displacement(
        self, cart_displacement: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Convert a Cartesian displacement into fractional coordinates.

        Parameters
        ----------
        cart_displacement
            (Å) Array with shape (N,3) where N is the number of atoms.

        Returns
        -------
        :
            (fractional) Array with shape (N,3).
        """
        verify_ndarray_shape("cart_displacement", cart_displacement, (None, 3))
        displacement = (cart_displacement) @ np.linalg.inv(self.lattice)
        return apply_pbc_displacement(displacement)

    def get_frac_direction(
        self, cart_direction: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Convert a Cartesian direction into fractional coordinates.

        Parameters
        ----------
        cart_direction
            (Å) Array with shape (3,).

        Returns
        -------
        :
            (fractional) Array with shape (3,).
        """
        verify_ndarray_shape("direction", cart_direction, (3,))
        displacement = np.array([cart_direction]) @ np.linalg.inv(self.lattice)
        return apply_pbc_displacement(displacement[0])

    def get_atom_indexes(self, atom_symbols: str | list[str]) -> list[int]:
        """Return atom indexes with matching symbols.

        Parameters
        ----------
        atom_symbols
            If integer or list of integers, specifies atom indexes. If string or list
            of strings, specifies atom symbols. Mixtures of indexes and symbols are
            allowed.
        """
        symbols = [ATOM_SYMBOLS[number] for number in self._atomic_numbers]
        indexes = []
        if isinstance(atom_symbols, str):
            atom_symbols = [atom_symbols]
        try:
            for index, symbol in enumerate(symbols):
                if symbol in atom_symbols:
                    indexes.append(index)
        except TypeError as err:
            raise get_type_error("atom_symbols", atom_symbols, "list") from err
        return indexes



"""Utility functions for spectra."""

import numpy as np
from numpy.typing import NDArray

import scipy
import scipy.fftpack

from ramannoodle.exceptions import verify_ndarray_shape, verify_ndarray, get_type_error


def convolve_spectrum(
    wavenumbers: NDArray[np.float64],
    intensities: NDArray[np.float64],
    function: str = "gaussian",
    width: float = 5,
    out_wavenumbers: NDArray[np.float64] | None = None,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    r"""Convolve a spectrum, producing a smoothing effect.

    Parameters
    ----------
    wavenumbers
        (cm\ :sup:`-1`) Array with shape (M,).
    intensities
        (arbitrary units) Array with shape (M,).
    function
        Supports ``"gaussian"`` or ``"lorentzian"``.
    width
        (cm\ :sup:`-1`)
    out_wavenumbers
        (cm\ :sup:`-1`) Array with shape (L,) where L is arbitrary. If ``None``,
        ``out_wavenumbers`` is determined automatically.

    Returns
    -------
    :
        0.  wavenumbers (``out_wavenumbers``) -- (cm\ :sup:`-1`) Array with shape (L,).
        #.  intensities -- (arbitrary units) Array with shape (L,).

    """
    if out_wavenumbers is None:
        min_wavenumber = np.min(wavenumbers) - 100
        max_wavenumber = np.max(wavenumbers) + 100
        num_samples = int(np.rint(max_wavenumber - min_wavenumber))
        out_wavenumbers = np.linspace(min_wavenumber, max_wavenumber, num_samples)
    verify_ndarray_shape("out_wavenumbers", out_wavenumbers, (None,))
    verify_ndarray_shape("wavenumbers", wavenumbers, (None,))
    verify_ndarray_shape("intensities", intensities, (len(wavenumbers),))
    try:
        if width <= 0:
            raise ValueError(f"invalid width: {width} <= 0")
    except TypeError as exc:
        raise get_type_error("width", width, "float") from exc
    verify_ndarray("out_wavenumbers", out_wavenumbers)

    convolved_intensities = out_wavenumbers * 0
    for wavenumber, intensity in zip(wavenumbers, intensities):
        factor = 0
        if function == "gaussian":
            factor = (
                (1 / width)
                * (1 / np.sqrt(2 * np.pi))
                * np.exp(-((wavenumber - out_wavenumbers) ** 2) / (2 * width**2))
            )
        elif function == "lorentzian":
            factor = (1 / np.pi) * (
                0.5 * width / ((wavenumber - out_wavenumbers) ** 2 + (0.5 * width) ** 2)
            )
        else:
            raise ValueError(f"unsupported convolution type: {function}")
        convolved_intensities += factor * intensity
    return (out_wavenumbers, convolved_intensities)


def _calc_autocorrelation(signal: NDArray[np.float64]) -> NDArray[np.float64]:
    """Calculate the positive autocorrelation of a signal.

    Parameters
    ----------
    signal
        Array with shape (S,) where S is the number of samples.

    Returns
    -------
    :
        Array with shape (S,])
    """
    verify_ndarray_shape("signal", signal, (None,))
    autocorrelation = scipy.signal.correlate(signal, signal, "full")
    autocorrelation = autocorrelation[(len(autocorrelation) - 1) // 2 :]
    return autocorrelation


def calc_signal_spectrum(
    signal: NDArray[np.float64], sampling_rate: float
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    r"""Calculate a signal's spectrum.

    The spectrum is  defined as the positive-frequency Fourier transform of the
    signal's autocorrelation.

    Parameters
    ----------
    signal
        Array with shape (S,) where S is the number of samples.
    sampling_rate
        (fs)

    Returns
    -------
    :
        0.  wavenumbers -- (cm\ :sup:`-1`) Array with shape (ceiling(S / 2),).
        #.  intensities -- (arbitrary units) Array with shape (ceiling(S / 2),).

    """
    autocorrelation = _calc_autocorrelation(signal)
    wavenumbers = (
        scipy.fftpack.fftfreq(autocorrelation.size, sampling_rate)
        * 33.35640951981521
        * 1e3
    )
    intensities = np.real(scipy.fftpack.fft(autocorrelation))
    return wavenumbers[wavenumbers >= 0], intensities[wavenumbers >= 0]



"""Abstract classes."""

from abc import ABC, abstractmethod


import numpy as np
from numpy.typing import NDArray


class PolarizabilityModel(ABC):  # pylint: disable=too-few-public-methods
    """Abstract polarizability model."""

    @abstractmethod
    def calc_polarizabilities(
        self, positions_batch: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Return estimated polarizabilities for a batch of fractional positions.

        Parameters
        ----------
        positions_batch
            (fractional) Array with shape (S,N,3) where S is the number of samples
            and N is the number of atoms.

        Returns
        -------
        :
            Array with shape (S,3,3).
        """


class RamanSpectrum(ABC):  # pylint: disable=too-few-public-methods
    """Abstract class for Raman spectra."""

    @abstractmethod
    def measure(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        orientation: str | NDArray[np.float64] = "polycrystalline",
        laser_correction: bool = False,
        laser_wavelength: float = 522,
        bose_einstein_correction: bool = False,
        temperature: float = 300,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""Calculate and return a raw Raman spectrum.

        Parameters
        ----------
        orientation
            Supports ``"polycrystalline"``. Future versions will support arbitrary
            orientations.
        laser_correction
            If ``True``, applies laser-wavelength-dependent intensity correction.
        laser_wavelength
            (nm) Ignored if ``laser_correction == False``.
        bose_einstein_correction
            If ``True``, applies temperature-dependent Bose Einstein correction.
        temperature
            (K) Ignored if ``bose_einstein_correction == False``.

        Returns
        -------
        :
            0. wavenumbers -- (cm\ :sup:`-1`) Array with shape (M,).

            #. intensities -- (arbitrary units) Array with shape (M,).

        """


class Dynamics(ABC):  # pylint: disable=too-few-public-methods
    """Abstract class for atomic dynamics."""

    @abstractmethod
    def get_raman_spectrum(
        self, polarizability_model: PolarizabilityModel
    ) -> RamanSpectrum:
        """Calculate a raman spectrum using a polarizability model.

        Parameters
        ----------
        polarizability_model
            Must be compatible with the dynamics.
        """



"""Functions for interacting with vasprun.xml files."""

from typing import TextIO
from pathlib import Path
from xml.etree.ElementTree import Element
import defusedxml.ElementTree as ET

import numpy as np
from numpy.typing import NDArray

from ramannoodle.io._utils import pathify, _read_polarizability_dataset
from ramannoodle.exceptions import InvalidFileException, UserError
from ramannoodle.constants import ATOMIC_WEIGHTS, ATOMIC_NUMBERS
from ramannoodle.dynamics._phonon import Phonons
from ramannoodle.dynamics._trajectory import Trajectory
from ramannoodle.structure._reference import ReferenceStructure

try:
    from ramannoodle.dataset.torch._dataset import PolarizabilityDataset
except UserError:
    pass


def _get_root_element(file: TextIO) -> Element:
    try:
        root = ET.parse(file).getroot()
        assert isinstance(root, Element)
        return root
    except ET.ParseError as exc:
        raise InvalidFileException("root xml element could not be found") from exc


def _parse_atomic_symbols(root: Element) -> list[str]:
    """Parse atomic symbols from a vasprun.xml Element.

    Raises
    ------
    InvalidFileException

    """
    set_element = root.find("./atominfo/array/set")
    if set_element is None:
        raise InvalidFileException("atomic symbols not found")

    atomic_symbols = []
    for child in set_element:
        atomic_symbol = child[0].text
        if atomic_symbol is None:
            raise InvalidFileException("child has no text")
        atomic_symbols.append(atomic_symbol.strip())
    return atomic_symbols


def _parse_positions(structure_varray: Element) -> NDArray[np.float64]:
    """Parse atomic fractional positions from a vasprun.xml Element.

    This function reads the positions from the initial structure.

    Raises
    ------
    InvalidFileException
    """
    positions = []
    for child in structure_varray:
        text = child.text
        if text is None:
            raise InvalidFileException("varray child text not found")
        positions.append([float(i) for i in text.split()])
    return np.array(positions)


def _parse_polarizability(root: Element) -> NDArray[np.float64]:
    """Parse polarizability from a vasprun.xml Element.

    In actuality, we read the macroscopic dielectric tensor including local field
    effects.

    Raises
    ------
    InvalidFileException
    """
    polarizability_element = root.find("./calculation/varray[@name='dielectric_dft']")
    if polarizability_element is None:
        raise InvalidFileException("polarizability not found")
    polarizability = []
    for child in polarizability_element:
        text = child.text
        if text is None:
            raise InvalidFileException("varray child text not found")
        polarizability.append([float(i) for i in text.split()])
    return np.array(polarizability)


def _parse_lattice(root: Element) -> NDArray[np.float64]:
    """Parse all three lattice vectors (in angstroms) from a vasprun.xml Element.

    Raises
    ------
    InvalidFileException
    """
    varray_element = root.find(
        "./structure[@name='initialpos']/crystal/varray[@name='basis']"
    )
    if varray_element is None:
        raise InvalidFileException("lattice not found")

    lattice = []
    for child in varray_element:
        text = child.text
        if text is None:
            raise InvalidFileException("varray child text not found")
        lattice.append([float(i) for i in text.split()])
    return np.array(lattice)


def read_positions_and_polarizability(
    filepath: str | Path,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Read fractional positions and polarizability from a vasprun.xml file.

    The polarizability returned by VASP is, in fact, a dielectric tensor. However,
    this is inconsequential to the calculation of Raman spectra.

    Parameters
    ----------
    filepath

    Returns
    -------
    :
        0.  positions -- (fractional) Array with shape (N,3) where N is the number of
            atoms.
        1.  polarizability -- Array with shape (3,3).

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        root = _get_root_element(file)
        structure_varray = root.find("./structure[@name='initialpos']/varray")
        if structure_varray is None:
            raise InvalidFileException("initial positions not found")
        positions = _parse_positions(structure_varray)
        polarizability = _parse_polarizability(root)
        return positions, polarizability


def read_structure_and_polarizability(
    filepath: str | Path,
) -> tuple[NDArray[np.float64], list[int], NDArray[np.float64], NDArray[np.float64]]:
    """Read lattice, positions, atomic numbers, and polarizability from a vasprun.xml.

    The polarizability returned by VASP is, in fact, a dielectric tensor. However,
    this is inconsequential to the calculation of Raman spectra.

    Parameters
    ----------
    filepath

    Returns
    -------
    :
        0.  lattice -- (Å) Array with shape (3,3).
        1.  atomic numbers -- List of length N where N is the number of atoms.
        2.  positions -- (fractional) Array with shape (N,3) where N is the number of
            atoms.
        3.  polarizability -- Array with shape (3,3).

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        root = _get_root_element(file)
        atomic_symbols = _parse_atomic_symbols(root)
        atomic_numbers = [ATOMIC_NUMBERS[symbol] for symbol in atomic_symbols]
        lattice = _parse_lattice(root)
        structure_varray = root.find("./structure[@name='initialpos']/varray")
        if structure_varray is None:
            raise InvalidFileException("initial positions not found")
        positions = _parse_positions(structure_varray)
        polarizability = _parse_polarizability(root)
        return lattice, atomic_numbers, positions, polarizability


def read_polarizability_dataset(
    filepaths: str | Path | list[str] | list[Path],
) -> "PolarizabilityDataset":
    """Read polarizability dataset from OUTCAR files.

    Parameters
    ----------
    filepaths

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    IncompatibleFileException
        File is incompatible with the dataset.
    """
    return _read_polarizability_dataset(filepaths, read_structure_and_polarizability)


def read_positions(filepath: str | Path) -> NDArray[np.float64]:
    """Read fractional positions from a vasprun.xml file.

    Parameters
    ----------
    filepath

    Returns
    -------
    :
        (fractional) Array with shape (N,3) where N is the number of atoms.

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        root = _get_root_element(file)
        structure_varray = root.find("./structure[@name='initialpos']/varray")
        if structure_varray is None:
            raise InvalidFileException("initial positions not found")
        positions = _parse_positions(structure_varray)
        return positions


def read_ref_structure(filepath: str | Path) -> ReferenceStructure:
    """Read reference structure from a vasprun.xml file.

    If the file contains multiple structures (such as those generated by a
    molecular dynamics run), the initial structure will be considered the reference
    structure.

    Parameters
    ----------
    filepath

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    SymmetryException
        Structural symmetry determination failed.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        root = _get_root_element(file)
        atomic_symbols = _parse_atomic_symbols(root)
        atomic_numbers = [ATOMIC_NUMBERS[symbol] for symbol in atomic_symbols]
        lattice = _parse_lattice(root)
        structure_varray = root.find("./structure[@name='initialpos']/varray")
        if structure_varray is None:
            raise InvalidFileException("initial positions not found")
        positions = _parse_positions(structure_varray)
        return ReferenceStructure(atomic_numbers, lattice, positions)


def _parse_timestep(root: Element) -> float:
    """Parse timestep in fs from a vasprun.xml Element.

    Raises
    ------
    InvalidFileException
    """
    element = root.find("./parameters/separator[@name='ionic']/i/[@name='POTIM']")
    if element is None:
        raise InvalidFileException("timestep not found")
    text = element.text
    if text is None:
        raise InvalidFileException("potim element has no text")
    return float(text.strip())


def read_trajectory(filepath: str | Path) -> Trajectory:
    """Read molecular dynamics trajectory from a vasprun.xml file.

    Parameters
    ----------
    filepath

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        root = _get_root_element(file)

        structures = root.iterfind("structure")
        positions_ts = []
        for structure in structures:
            if "name" in structure.attrib:  # skip named structures
                continue
            structure_varray = structure.find("varray")
            if structure_varray is None:
                raise InvalidFileException("structure varray not found")
            positions_ts.append(_parse_positions(structure_varray))
        if len(positions_ts) == 0:
            raise InvalidFileException("no trajectory found")

        timestep = _parse_timestep(root)

        return Trajectory(np.array(positions_ts), timestep)


def _parse_eigenvalues(root: Element) -> NDArray[np.float64]:
    """Parse eigenvalues in cm-1 from a vasprun.xml Element.

    Raises
    ------
    InvalidFileException
    """
    element = root.find("./calculation/dynmat/v[@name='eigenvalues']")
    if element is None:
        raise InvalidFileException("eigenvalues not found")
    if element.text is None:
        raise InvalidFileException("eigenvalues has no text")
    eigenvalues = np.array([float(i) for i in element.text.split()])
    eigenvalues = np.sqrt(np.abs(-eigenvalues)) * 1 / 0.0299792458
    eigenvalues *= np.sign(eigenvalues)
    return eigenvalues


def _parse_eigenvectors(root: Element) -> NDArray[np.float64]:
    """Parse eigenvectors in cm-1 from a vasprun.xml Element.

    Raises
    ------
    InvalidFileException
    """
    lattice = _parse_lattice(root)
    num_atoms = len(_parse_atomic_symbols(root))
    element = root.find("./calculation/dynmat/varray[@name='eigenvectors']")
    if element is None:
        raise InvalidFileException("eigenvectors not found")

    eigenvectors = []
    for v in element:
        if v.text is None:
            raise InvalidFileException("eigenvector element has no text")
        cart_eigenvector = np.array([float(i) for i in v.text.split()])
        cart_eigenvector = cart_eigenvector.reshape((num_atoms, 3))
        eigenvectors.append(cart_eigenvector @ np.linalg.inv(lattice))

    return np.array(eigenvectors)


def read_phonons(filepath: str | Path) -> Phonons:
    """Read phonons from a vasprun.xml file.

    Parameters
    ----------
    filepath

    Returns
    -------
    :

    Raises
    ------
    FileNotFoundError
        File not found.
    InvalidFileException
        Invalid file.
    """
    filepath = pathify(filepath)
    with open(filepath, "r", encoding="utf-8") as file:
        root = _get_root_element(file)

        atomic_symbols = _parse_atomic_symbols(root)
        atomic_weights = np.array([ATOMIC_WEIGHTS[symbol] for symbol in atomic_symbols])
        structure_varray = root.find("./structure[@name='initialpos']/varray")
        if structure_varray is None:
            raise InvalidFileException("initial positions not found")
        ref_positions = _parse_positions(structure_varray)

        wavenumbers = _parse_eigenvalues(root)
        eigenvectors = _parse_eigenvectors(root)
        displacements = eigenvectors / np.sqrt(atomic_weights)[:, np.newaxis]

        return Phonons(ref_positions, wavenumbers, displacements)
